#include <AnaEeh.hh>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/LCFloatVec.h>
#include <EVENT/MCParticle.h>
#include <EVENT/ReconstructedParticle.h>
#include <IMPL/MCParticleImpl.h>
#include <values.h>
#include <string>
#include <iostream>
#include <fstream>
#include <EVENT/LCFloatVec.h>
#include <EVENT/LCParameters.h>
#include <stdexcept>
#include <TFile.h> 
#include <TTree.h>
#include <TVector3.h>
#include <TRandom.h>
#include <Rtypes.h> 
#include <sstream>		
#include <cmath>
#include <vector>
#include <TMath.h>
#include "TLorentzVector.h"
#include <stdexcept>
#include <cstdlib>

using namespace std;

const float sqrts = 250.0; 	//GeV

AnaEeh a_AnaEeh;
AnaEeh::AnaEeh()
	: Processor("AnaEeh"),
	_output(0)
{
	_description = "llH analysis" ;

	_treeFileName="llH.root";
	registerProcessorParameter( "TreeOutputFile" , 
			"The name of the file to which the ROOT tree will be written" ,
			_treeFileName ,
			_treeFileName);

	_colName="MCParticle";
	registerProcessorParameter( "MCObjects" ,
			"The name of the PFOs" ,
			_colName ,
			_colName);

	_treeName="MCPart";
	registerProcessorParameter( "TreeName" , 
			"The name of the ROOT tree" ,
			_treeName ,
			_treeName);

	_leptonID = 13;
	registerProcessorParameter( "LeptonIDTag" ,
                        "Lepton ID that will be used in this analysis." ,
                        _leptonID ,
                        _leptonID);
        _fileID = 1;
        registerProcessorParameter( "fileIDTag" ,
                        "File ID that will be used in this analysis." ,
                        _fileID ,
                        _fileID);
	
	_overwrite=0;
	registerProcessorParameter( "OverwriteFile" , 
			"If zero an already existing file will not be overwritten." ,
			_overwrite ,
			_overwrite);

}

void AnaEeh::init() {

	printParameters();
	TFile *tree_file=new TFile(_treeFileName.c_str(),(_overwrite ? "RECREATE" : "UPDATE"));

	if (!tree_file->IsOpen()) {
		delete tree_file;
		tree_file=new TFile(_treeFileName.c_str(),"NEW");
	}

	_outputTree = new TTree(_treeName.c_str(),_treeName.c_str());
	_outputTree->SetAutoSave(32*1024*1024);  // autosave every 32MB
	_outputTree->Branch("EventNr", &_eventNr, "EventNr/I");
	_outputTree->Branch("Num", &_Num, "Num/I");
	_outputTree->Branch("FileNr", &_FileNr, "FileNr/I");
	_outputTree->Branch("nHDleptonM", &nHDleptonM, "nHDleptonM/I");
	_outputTree->Branch("nHDleptonP", &nHDleptonP, "nHDleptonP/I");
	_outputTree->Branch("nHDGamma", &nHDGamma, "nHDGamma/I");
	_outputTree->Branch("nHDQuark", &nHDQuark, "nHDQuark/I");
	_outputTree->Branch("nHDGluon", &nHDGluon, "nHDGluon/I");
	_outputTree->Branch("nHDOther", &nHDOther, "nHDOther/I");
	_outputTree->Branch("nNePFO", &nNePFO, "nNePFO/I");
	_outputTree->Branch("nChPFO", &nChPFO, "nChPFO/I");
	_outputTree->Branch("nPhotonPFO", &nPhotonPFO, "nPhotonPFO/I");
	_outputTree->Branch("nMuonMPFO", &nMuonMPFO, "nMuonMPFO/I");
	_outputTree->Branch("nMuonPPFO", &nMuonPPFO, "nMuonPPFO/I");
	_outputTree->Branch("nEleMPFO", &nEleMPFO, "nEleMPFO/I");
	_outputTree->Branch("nElePPFO", &nElePPFO, "nElePPFO/I");
	_outputTree->Branch("nHadPFO", &nHadPFO, "nHadPFO/I");
	_outputTree->Branch("nBremPhM", &nBremPhM, "nBremPhM/I");
	_outputTree->Branch("nBremPhP", &nBremPhP, "nBremPhP/I");
	_outputTree->Branch("InvMass", &_InvMass, "InvMass/F");
	_outputTree->Branch("InvMassH", &_InvMassH, "InvMassH/F");
	_outputTree->Branch("RecoilMass", &_RecoilMass, "RecoilMass/F");
	_outputTree->Branch("InvMassABC", &InvMass_ABC, "InvMassABC/F");
	_outputTree->Branch("InvMassHABC", &InvMassH_ABC, "InvMassHABC/F");
	_outputTree->Branch("RecoilMassABC", &RecoilMass_ABC, "RecoilMassABC/F");
	
	_FileNr = _fileID;
	_Num = 0;
}

void AnaEeh::processEvent( LCEvent * evtP ) 
{		

	if (evtP) 								
	{		
		try 	
		{ 
			cout << "start:" << _Num << endl;
			_eventNr=evtP->getEventNumber();
			LCCollection* col_MCP = evtP->getCollection( _colName ) ;
			LCCollection* col_RecoP = evtP->getCollection( "ArborPFOs" );
			int _nMCP=col_MCP->getNumberOfElements();
			int _nRecoP = col_RecoP->getNumberOfElements();	
			TLorentzVector P_P, P_M, P_Ph;
			TLorentzVector P_T(0, 0, 0, sqrts);
			TLorentzVector P_tot(0, 0, 0, 0);
			cout << _nMCP << " : " << _nRecoP << endl;		
			std::vector<MCParticle*> HDleptonM;
			std::vector<MCParticle*> HDleptonP;
			std::vector<MCParticle*> HDGamma;
			std::vector<MCParticle*> HDQuark;
			std::vector<MCParticle*> HDGluon;
			std::vector<MCParticle*> HDOther;
			std::vector<ReconstructedParticle*> NePFO;
			std::vector<ReconstructedParticle*> ChPFO;
			std::vector<ReconstructedParticle*> PhotonPFO;
			std::vector<ReconstructedParticle*> MuonMPFO;
			std::vector<ReconstructedParticle*> MuonPPFO;
			std::vector<ReconstructedParticle*> EleMPFO;
			std::vector<ReconstructedParticle*> ElePPFO;
			std::vector<ReconstructedParticle*> HadPFO;
			std::vector<ReconstructedParticle*> BremPhM;
			std::vector<ReconstructedParticle*> BremPhP;
			HDleptonM.clear();
			HDleptonP.clear();
			HDGamma.clear();
			HDQuark.clear();
			HDGluon.clear();
			HDOther.clear();
			NePFO.clear();
			ChPFO.clear();
			PhotonPFO.clear();
			MuonMPFO.clear();
			MuonPPFO.clear();
			EleMPFO.clear();
			ElePPFO.clear();
			HadPFO.clear();
			BremPhM.clear();
			BremPhP.clear();
			nHDleptonM = 0;
			nHDleptonP = 0;
			nHDGamma = 0;
			nHDQuark = 0;
			nHDGluon = 0;
			nHDOther = 0;
			nNePFO = 0;
			nChPFO = 0;
			nPhotonPFO = 0;
			nMuonMPFO = 0;
			nMuonPPFO = 0;
			nEleMPFO = 0;
			nElePPFO = 0;
			nHadPFO = 0;
			nBremPhM = 0;
			nBremPhP = 0;
			_InvMass = -1;
			_InvMassH = -1;
			_RecoilMass = -1;
			InvMass_ABC = -1;
			InvMassH_ABC = -1;
			RecoilMass_ABC = -1;

			for(int i = 0; i < _nMCP; i++)
			{
				MCParticle *a1_MCP = dynamic_cast<EVENT::MCParticle *>(col_MCP->getElementAt(i));
				int tmpPID = a1_MCP->getPDG();
				int NParent = a1_MCP->getParents().size();
				int NDaughter = a1_MCP->getDaughters().size();
				TLorentzVector _Ph(0,0,0,0);
				if(tmpPID == 25 && NDaughter > 1 && NParent !=0 ) //Higgs
				{
					
					for(int i1 = 0; i1 < NDaughter; i1++)  
					{
						MCParticle *b_MCP = a1_MCP->getDaughters()[i1];
						int _HdaughterPID = b_MCP->getPDG();

						if(abs(_HdaughterPID) == 11 || abs(_HdaughterPID) == 13)
						{
							if(_HdaughterPID < 0) HDleptonP.push_back(b_MCP);
							else HDleptonM.push_back(b_MCP);
						}
						else if(abs(_HdaughterPID) == 22)
						{
							HDGamma.push_back(b_MCP);
						}
						else if(abs(_HdaughterPID) < 6)
						{
							HDQuark.push_back(b_MCP);
						}
						else if(abs(_HdaughterPID) == 21)
						{
							HDGluon.push_back(b_MCP);
						}
						else HDOther.push_back(b_MCP);

					}
				}
				
				
			}
			
			
			nHDleptonM = HDleptonM.size();
			nHDleptonP = HDleptonP.size();
			nHDGamma = HDGamma.size();
			nHDQuark = HDQuark.size();
			nHDGluon = HDGluon.size();
			nHDOther = HDOther.size();

			for(int i0 = 0; i0 < _nRecoP; i0++)
			{
				ReconstructedParticle *a_RecoP = dynamic_cast<EVENT::ReconstructedParticle *>(col_RecoP->getElementAt(i0));
				TLorentzVector P_PFO(0,0,0,0);
				P_PFO[3] = a_RecoP->getEnergy();
				P_PFO[0] = a_RecoP->getMomentum()[0];
				P_PFO[1] = a_RecoP->getMomentum()[1];
				P_PFO[2] = a_RecoP->getMomentum()[2];
				P_tot += P_PFO;
				int Charge = a_RecoP->getCharge();
				int PID = a_RecoP->getType();
				if (Charge == 0)
				{
					NePFO.push_back(a_RecoP);
					if(PID == 22) PhotonPFO.push_back(a_RecoP);
				}
				else
				{
					ChPFO.push_back(a_RecoP);
					if(PID == 13) MuonMPFO.push_back(a_RecoP);
					else if (PID == -13) MuonPPFO.push_back(a_RecoP);
					else if (PID == 11) EleMPFO.push_back(a_RecoP);
					else if (PID == -11) ElePPFO.push_back(a_RecoP);
					else HadPFO.push_back(a_RecoP);
				}
			}
			
			nNePFO = NePFO.size();
			nChPFO = ChPFO.size();
			nPhotonPFO = PhotonPFO.size();
			nMuonMPFO = MuonMPFO.size();
			nMuonPPFO = MuonPPFO.size();
			nEleMPFO = EleMPFO.size();
			nElePPFO = ElePPFO.size();
			nHadPFO = HadPFO.size();

			float MinZThrDis = 1.0E10;

			int indexM = -1;
			int indexP = -1;
			if(_leptonID == 13 && nMuonMPFO >= 0 && nMuonPPFO >=0)
			{
				
				for(int i1 = 0; i1 < nMuonMPFO; i1++)
				{
					ReconstructedParticle *a_RecoMP = MuonMPFO[i1];
					P_M[3] = a_RecoMP->getEnergy();
					P_M[0] = a_RecoMP->getMomentum()[0];
					P_M[1] = a_RecoMP->getMomentum()[1];
					P_M[2] = a_RecoMP->getMomentum()[2];
	
					for(int i2 = 0; i2 < nMuonPPFO; i2++)
					{
						ReconstructedParticle *a_RecoPP = MuonPPFO[i2];
						P_P[3] = a_RecoPP->getEnergy();
						P_P[0] = a_RecoPP->getMomentum()[0];
						P_P[1] = a_RecoPP->getMomentum()[1];
						P_P[2] = a_RecoPP->getMomentum()[2];
						float currInvMass = (P_P + P_M).M();
						float currRecoilMass = (P_T - P_P - P_M).M();
						if(fabs(currInvMass - 91.2) < MinZThrDis)
						{
							_InvMass = currInvMass; 
							_RecoilMass = currRecoilMass;
							_InvMassH = (P_tot- P_P - P_M).M();
							MinZThrDis = fabs(currInvMass - 91.2);
							indexM = i1;
							indexP = i2;  
						}
					} 
					
				}
					
				if(indexM >= 0 && indexP >= 0)
				{
				ReconstructedParticle *a_RecoMP = MuonMPFO[indexM];
				ReconstructedParticle *a_RecoPP = MuonPPFO[indexP];
				P_M[3] = a_RecoMP->getEnergy();
				P_M[0] = a_RecoMP->getMomentum()[0];
				P_M[1] = a_RecoMP->getMomentum()[1];
				P_M[2] = a_RecoMP->getMomentum()[2];
				P_P[3] = a_RecoPP->getEnergy();
				P_P[0] = a_RecoPP->getMomentum()[0];
				P_P[1] = a_RecoPP->getMomentum()[1];
				P_P[2] = a_RecoPP->getMomentum()[2];
				float Theta_M = P_M.Theta();
				float Phi_M = P_M.Phi();
				float Pt_M = P_M.Perp();
				float Theta_P = P_P.Theta();
				float Phi_P = P_P.Phi();
				float Pt_P = P_P.Perp();
				float cutPM;
				float cutPP;
				if (fabs(cos(Theta_M)) < fabs(cos(atan(1808/2350.))) )
				{
					cutPM = asin(0.9/(Pt_M/(0.3*3.5)));
				}
				else cutPM = asin(2.35*Pt_M/P_M[2]/(Pt_M/(0.3*3.5)));
				if (fabs(cos(Theta_P)) < fabs(cos(atan(1808/2350.))) )
				{
					cutPP = asin(0.9/(Pt_P/(0.3*3.5)));
				}
				else cutPP = asin(2.35*Pt_P/P_P[2]/(Pt_P/(0.3*3.5)));	
				for(int k = 0; k < nPhotonPFO; k++)
				{
					ReconstructedParticle *a_RecoPhP = PhotonPFO[k];
					P_Ph[3] = a_RecoPhP->getEnergy();
					P_Ph[0] = a_RecoPhP->getMomentum()[0];
					P_Ph[1] = a_RecoPhP->getMomentum()[1];
					P_Ph[2] = a_RecoPhP->getMomentum()[2];
					float Theta_Ph = P_Ph.Theta();
					float Phi_Ph = P_Ph.Phi();
					if(fabs(Theta_Ph-Theta_M) < 0.005 && Phi_Ph-Phi_M >= 0 && Phi_Ph-Phi_M <= cutPM)
					{
						BremPhM.push_back(a_RecoPhP);
						P_M = P_M + P_Ph;
					}
					else if(fabs(Theta_Ph-Theta_P) < 0.005 && Phi_P-Phi_Ph >= 0 && Phi_P-Phi_Ph <= cutPP)
					{
						BremPhP.push_back(a_RecoPhP);
						P_P = P_P + P_Ph;
					}
										
				}
				
				InvMass_ABC = (P_P + P_M).M();
				RecoilMass_ABC = (P_T - P_P - P_M).M();
				InvMassH_ABC = (P_tot - P_P - P_M).M();
				}
				else
				{
				InvMass_ABC = _InvMass;
				RecoilMass_ABC = _RecoilMass;
				InvMassH_ABC = _InvMassH;
				}
								
				
			}
			else if (_leptonID == 11 && nEleMPFO >= 0 && nElePPFO >=0)
			{
				for(int i1 = 0; i1 < nEleMPFO; i1++)
				{
					ReconstructedParticle *a_RecoMP = EleMPFO[i1];
					P_M[3] = a_RecoMP->getEnergy();
					P_M[0] = a_RecoMP->getMomentum()[0];
					P_M[1] = a_RecoMP->getMomentum()[1];
					P_M[2] = a_RecoMP->getMomentum()[2];
	
					for(int i2 = 0; i2 < nElePPFO; i2++)
					{
						ReconstructedParticle *a_RecoPP = ElePPFO[i2];
						P_P[3] = a_RecoPP->getEnergy();
						P_P[0] = a_RecoPP->getMomentum()[0];
						P_P[1] = a_RecoPP->getMomentum()[1];
						P_P[2] = a_RecoPP->getMomentum()[2];
						float currInvMass = (P_P + P_M).M();
						float currRecoilMass = (P_T - P_P - P_M).M();
						if(fabs(currInvMass - 91.2) < MinZThrDis)
						{
							_InvMass = currInvMass; 
							_RecoilMass = currRecoilMass;
							_InvMassH = (P_tot- P_P - P_M).M();
							MinZThrDis = fabs(currInvMass - 91.2);
							indexM = i1;
							indexP = i2; 
						}
					} 
					
				}	
				if(indexM >= 0 && indexP >= 0)
				{			
				ReconstructedParticle *a_RecoMP = EleMPFO[indexM];
				ReconstructedParticle *a_RecoPP = ElePPFO[indexP];
				P_M[3] = a_RecoMP->getEnergy();
				P_M[0] = a_RecoMP->getMomentum()[0];
				P_M[1] = a_RecoMP->getMomentum()[1];
				P_M[2] = a_RecoMP->getMomentum()[2];
				P_P[3] = a_RecoPP->getEnergy();
				P_P[0] = a_RecoPP->getMomentum()[0];
				P_P[1] = a_RecoPP->getMomentum()[1];
				P_P[2] = a_RecoPP->getMomentum()[2];
				float Theta_M = P_M.Theta();
				float Phi_M = P_M.Phi();
				float Pt_M = P_M.Perp();
				float Theta_P = P_P.Theta();
				float Phi_P = P_P.Phi();
				float Pt_P = P_P.Perp();
				float cutPM;
				float cutPP;
				if (fabs(cos(Theta_M)) < fabs(cos(atan(1808/2350.))) )
				{
					cutPM = asin(0.9/(Pt_M/(0.3*3.5)));
				}
				else cutPM = asin(2.35*Pt_M/P_M[2]/(Pt_M/(0.3*3.5)));
				if (fabs(cos(Theta_P)) < fabs(cos(atan(1808/2350.))) )
				{
					cutPP = asin(0.9/(Pt_P/(0.3*3.5)));
				}
				else cutPP = asin(2.35*Pt_P/P_P[2]/(Pt_P/(0.3*3.5)));	
				for(int k = 0; k < nPhotonPFO; k++)
				{
					ReconstructedParticle *a_RecoPhP = PhotonPFO[k];
					P_Ph[3] = a_RecoPhP->getEnergy();
					P_Ph[0] = a_RecoPhP->getMomentum()[0];
					P_Ph[1] = a_RecoPhP->getMomentum()[1];
					P_Ph[2] = a_RecoPhP->getMomentum()[2];
					float Theta_Ph = P_Ph.Theta();
					float Phi_Ph = P_Ph.Phi();
					if(fabs(Theta_Ph-Theta_M) < 0.005 && Phi_Ph-Phi_M >= 0 && Phi_Ph-Phi_M <= cutPM)
					{
						BremPhM.push_back(a_RecoPhP);
						P_M = P_M + P_Ph;
					}
					else if(fabs(Theta_Ph-Theta_P) < 0.005 && Phi_P-Phi_Ph >= 0 && Phi_P-Phi_Ph <= cutPP)
					{
						BremPhP.push_back(a_RecoPhP);
						P_P = P_P + P_Ph;
					}
										
				}
				
				InvMass_ABC = (P_P + P_M).M();
				RecoilMass_ABC = (P_T - P_P - P_M).M();
				InvMassH_ABC = (P_tot - P_P - P_M).M();
				}
				else
				{
				InvMass_ABC = _InvMass;
				RecoilMass_ABC = _RecoilMass;
				InvMassH_ABC = _InvMassH;
				}
			}
			else
			{
				_InvMass = -1;
				_RecoilMass = -1;
				_InvMassH = -1;
				InvMass_ABC = -1;
				RecoilMass_ABC = -1;
				InvMassH_ABC = -1;
			}
			nBremPhM = BremPhM.size();
			nBremPhP = BremPhP.size();
			
			
			
			_outputTree->Fill();
			_Num++;
		}		
		catch (lcio::DataNotAvailableException err) { }

	}  	  

}


void AnaEeh::end()
{

	if (_outputTree) {

		TFile *tree_file = _outputTree->GetCurrentFile();
		tree_file->Write();
		delete tree_file;
	}

}	


