#include <Trknum.hh>
#include "ArborTool.hh"
#include "ArborToolLCIO.hh"
#include "DetectorPos.hh"
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/ReconstructedParticle.h>
#include <EVENT/CalorimeterHit.h>
#include <EVENT/SimCalorimeterHit.h>
#include <values.h>
#include <string>
#include <iostream>
#include <cmath>
#include <vector>
#include <stdlib.h>
#include <EVENT/LCFloatVec.h>
#include <EVENT/LCParameters.h>
#include <EVENT/Cluster.h>
#include <EVENT/LCRelation.h>
#include <UTIL/CellIDDecoder.h>
#include <stdexcept>
#include <TFile.h> 
#include <TTree.h>
#include <TMath.h>
#include <Rtypes.h> 
#include <sstream>		
#include <TH1.h>
#include <TVector3.h>



using namespace std;

Trknum aTrknum ;
Trknum::Trknum()
	: Processor("Trknum"),
	_output(0)
{
	_description = "PFO Track num" ;

	_treeFileName="Trknum.root";
	registerProcessorParameter( "TreeOutputFile" , 
			"The name of the file to which the ROOT tree will be written" ,
			_treeFileName ,
			_treeFileName);

	_treeName="ArborPFO";
	registerProcessorParameter( "TreeName" , 
			"The name of the ROOT tree" ,
			_treeName ,
			_treeName);

	_overwrite=0;
	registerProcessorParameter( "OverwriteFile" , 
			"If zero an already existing file will not be overwritten." ,
			_overwrite ,
			_overwrite);

	_colName="ChargedCore";
	registerProcessorParameter( "InputArborCollection" ,
                        "Name of Reconstructed Arbor Collections" ,
			_colName ,
			_colName);
/*
        std::vector<std::string> inputArborClusters;
	inputArborClusters.push_back("EcalTrees");	//Trees & Bushes
	inputArborClusters.push_back("EcalBushes");
	inputArborClusters.push_back("HcalTrees");
	inputArborClusters.push_back("HcalBushes");
	registerProcessorParameter( "InputArborCluster" ,
                        "Name of Cluster Collections" ,
			_colName2 ,
			_colName2);
*/
}

void Trknum::init() {

	printParameters();

	TFile *tree_file=new TFile(_treeFileName.c_str(),(_overwrite ? "RECREATE" : "UPDATE"));

	if (!tree_file->IsOpen()) {
		delete tree_file;
		tree_file=new TFile(_treeFileName.c_str(),"NEW");
	}

	_outputTree = new TTree(_treeName.c_str(),_treeName.c_str());
	_outputTree->SetAutoSave(32*1024*1024);  // autosave every 32MB

	_outputPFO = new TTree("ChPFO", "ChPFO" );
	_outputPFO->SetAutoSave(32*1024*1024);  // autosave every 32MB
 
        _outputTree->Branch("EventNr", &_eventNr, "EventNr/I");
        _outputTree->Branch("Num", &num,"Num/I");
        _outputTree->Branch("ThetaMC", &_thetaMC, "ThetaMC/F");
        _outputTree->Branch("ThetaDisCh", &_thetaDisCh, "ThetaDisCh/F");
        _outputTree->Branch("ThetaDisMip", &_thetaDisMip, "ThetaDisMip/F");
        _outputTree->Branch("ThetaDisEm", &_thetaDisEm, "ThetaDisEm/F");
        _outputTree->Branch("ThetaDisHad", &_thetaDisHad, "ThetaDisHad/F");
        _outputTree->Branch("nCH", &_nCH, "nCH/I");
        _outputTree->Branch("nMIPs", &_nMIPs, "nMIPs/I");
        _outputTree->Branch("nEMs", &_nEMs, "nEMs/I");
        _outputTree->Branch("nHADs", &_nHADs, "nHADs/I");
        _outputTree->Branch("nMuPFOs", &_nMuPFO, "nMuPFOs/I");
        _outputTree->Branch("nEPFOs", &_nEPFO, "nEPFOs/I");
        _outputTree->Branch("nPiPFOs", &_nPiPFO, "nPiPFOs/I");
   
        _outputPFO->Branch("EventNr", &_eventNr, "EventNr/I");
	_outputPFO->Branch("Num", &num,"Num/I");
        _outputPFO->Branch("TrkEn", &_TrkEn, "TrkEn/F");
        _outputPFO->Branch("Type", &_type, "Type/I");
        _outputPFO->Branch("Charge", &_Charge, "Charge/I");
	_outputPFO->Branch("EClu", &_EClu, "EClu/F");
	_outputPFO->Branch("EE_Clu", &_EcalEn, "EE_Clu/F");
	_outputPFO->Branch("EH_Clu", &_HcalEn, "EH_Clu/F");
	_outputPFO->Branch("NH_ECAL", &_EcalNHit, "NH_ECAL/I");
	_outputPFO->Branch("NH_HCAL", &_HcalNHit, "NH_HCAL/I");
	_outputPFO->Branch("FD_ECAL", &_FD_ECAL, "FD_ECAL/F");
        _outputPFO->Branch("FD_HCAL", &_FD_HCAL, "FD_HCAL/F");
	_outputPFO->Branch("FD_all", &_FD_all, "FD_all/F");
	_outputPFO->Branch("FD", _FD, "FD[16]/F");	//Divide into 6 parts; 
	_outputPFO->Branch("NH", _NH, "NH[16]/I");
	_outputPFO->Branch("NL", _NL, "NL[16]/I");
	_outputPFO->Branch("NLEcal", &_NLEcal, "NLEcal/I");
	_outputPFO->Branch("NLHcal", &_NLHcal, "NLHcal/I");
	_outputPFO->Branch("MaxDisHel", &_MaxDisHel, "MaxDisHel/F");	//Maximal Distance from Hit to Helix
	_outputPFO->Branch("MinDisHel", &_MinDisHel, "MinDisHel/F");
	_outputPFO->Branch("minDepth", &_minDepth, "minDepth/F");
	_outputPFO->Branch("maxDepth", &_maxDepth, "maxDepth/F");
	_outputPFO->Branch("CluFlagID", &_CluFlagID, "CluFlagID/I");
	_outputPFO->Branch("graDepth", &_graDepth, "graDepth/F");
	_outputPFO->Branch("CluDepth", &_cluDepth, "CluDepth/F");
	_outputPFO->Branch("EECluL10", &_EEClu_L10, "EECluL10/F");
	_outputPFO->Branch("EECluR", &_EEClu_R, "EECluR/F");
	_outputPFO->Branch("EEClur", &_EEClu_r, "EEClur/F");
	_outputPFO->Branch("EEClup", &_EEClu_p, "EEClup/F");
	_outputPFO->Branch("NALEcal", &_AL_Ecal, "NALEcal/I");
	_outputPFO->Branch("NALHcal", &_AL_Hcal, "NALHcal/I");
	_outputPFO->Branch("rmsEcal", &_rms_Ecal, "rmsEcal/F");
	_outputPFO->Branch("rmsHcal", &_rms_Hcal, "rmsHcal/F");
	_outputPFO->Branch("rmsEcal2", &_rms_Ecal2, "rmsEcal2/F");
	_outputPFO->Branch("rmsHcal2", &_rms_Hcal2, "rmsHcal2/F");
	_outputPFO->Branch("avNHE", &_av_NHE, "avNHE/F");
	_outputPFO->Branch("avNHH", &_av_NHH, "avNHH/F");
	_outputPFO->Branch("dEdx",&_dEdx,"dEdx/F");
	
	_outputPFO->Branch("avDisHtoL", &_avDisHtoL, "avDisHtoL/F");
	_outputPFO->Branch("avEnDisHtoL", &_avEnDisHtoL, "avEnDisHtoL/F");
	_outputPFO->Branch("PhiBumpNum",&_nPhiBump,"PhiBumpNum/I");
	_outputPFO->Branch("FD_ECALF10", &_FD_ECALF10,"FD_ECALF10/F");
	_outputPFO->Branch("NH_ECALF10",&_NH_ECALF10,"NH_ECALF10/I");
        _outputPFO->Branch("FD_ECALL20", &_FD_ECALL20,"FD_ECALL20/F");
        _outputPFO->Branch("NH_ECALL20",&_NH_ECALL20,"NH_ECALL20/I");
	num = 0;
        
}



void Trknum::processEvent( LCEvent * evtP ) 
{		

	if (evtP) 								
	{	
		try 	
		{  
                   
                    int nChPFO = 0, nEM = 0, nMIP = 0, nHAD = 0;
                    _nMuPFO = 0;
                    _nEPFO = 0;
 		    _nPiPFO = 0;
                    _eventNr=evtP->getEventNumber();
	            LCCollection * MCP = evtP->getCollection("MCParticle");
                    for(int i0 = 0; i0 < MCP->getNumberOfElements(); i0++)
		    {
			MCParticle * a_MCP = dynamic_cast<MCParticle*>(MCP->getElementAt(i0));
			if( a_MCP->getParents().size() == 0 )
			{
				
				_MCPEn = a_MCP->getEnergy();                            
				_MCPP[0] = a_MCP->getMomentum()[0];
				_MCPP[1] = a_MCP->getMomentum()[1];
				_MCPP[2] = a_MCP->getMomentum()[2];
			}
		    }
                    _thetaMC = TMath::ACos(_MCPP[2]/sqrt(_MCPP[0]*_MCPP[0]+_MCPP[1]*_MCPP[1]+_MCPP[2]*_MCPP[2]));



                    cout << "Event:" << _eventNr << endl;
                    LCCollection * col_RecoP = evtP->getCollection("ArborPFOs");




                    for(int i0 = 0; i0 < col_RecoP->getNumberOfElements(); i0++)
		    {
			ReconstructedParticle *a_RecoP = dynamic_cast<EVENT::ReconstructedParticle *>(col_RecoP->getElementAt(i0));	
                        _Charge = a_RecoP->getCharge();
                        _type = a_RecoP->getType();
                        if (_Charge != 0)
                        {
                            nChPFO ++;
                            if(_type == 13 || _type == -13) nMIP++;
                            else if(_type == 11 || _type == -11) nEM++;
                            else nHAD++;
                            if(a_RecoP->getClusters().size() > 0)
                            {
				_EClu = 0;
                                TVector3 CluPos;
                                Cluster * currClu = a_RecoP->getClusters()[0];
				_EClu = currClu->getEnergy();
                                Track * a_trk = a_RecoP->getTracks()[0];
                                _TrkEn = a_RecoP->getEnergy();
                                
				_dEdx = a_trk->getdEdx();
                                CluPos = currClu->getPosition();
                                TVector3 IntDir = ClusterCoG(currClu)-CluPos;
        			const string ECALCellIDDecoder = "M:3,S-1:3,I:9,J:9,K-1:6";
        			CellIDDecoder<CalorimeterHit> idDecoder(ECALCellIDDecoder);
        			
        			
        			
        			_EcalEn = 0;
        			_HcalEn = 0;
        			_EcalNHit = 0;
        			_HcalNHit = 0;
        			_CluNHit = 0;
				_NH_ECALF10 = 0;
                                _EEClu_R = 0;
                                _EEClu_r = 0;
                                _EEClu_p = 0;
                                _EEClu_L10 = 0;
        			float currDepth = 0;
        			_maxDepth = -100;
        			_minDepth = 1E6;
        			_MaxDisHel = -1;   //maximal distance from Track to Helix
        			_MinDisHel = 1E10;

        			std::vector<CalorimeterHit*> Ecalhits;
        			std::vector<CalorimeterHit*> Hcalhits;
        			std::vector<CalorimeterHit*> allhits;
        			std::vector<CalorimeterHit*> EH_1;
        			std::vector<CalorimeterHit*> EH_2;
        			std::vector<CalorimeterHit*> EH_3;
        			std::vector<CalorimeterHit*> EH_4;
        			std::vector<CalorimeterHit*> EH_5;
        			std::vector<CalorimeterHit*> EH_6;
        			std::vector<CalorimeterHit*> HH_1;
        			std::vector<CalorimeterHit*> HH_2;
        			std::vector<CalorimeterHit*> HH_3;
        			std::vector<CalorimeterHit*> HH_4;
        			std::vector<CalorimeterHit*> HH_5;
        			std::vector<CalorimeterHit*> HH_6;
        			std::vector<CalorimeterHit*> HH_7;
        			std::vector<CalorimeterHit*> HH_8;
        			std::vector<CalorimeterHit*> HH_9;
        			std::vector<CalorimeterHit*> HH_0;
				std::vector<CalorimeterHit*> Ecalf10hits;	
				std::vector<CalorimeterHit*> Ecall20hits;			

        			allhits.clear();
        			Ecalhits.clear();
        			Hcalhits.clear();
        			EH_1.clear();
       				EH_2.clear();
        			EH_3.clear();
        			EH_4.clear();
       				EH_5.clear();
        			EH_6.clear();
        			HH_1.clear();
        			HH_2.clear();
        			HH_3.clear();
        			HH_4.clear();
        			HH_5.clear();
        			HH_6.clear();
        			HH_7.clear();
        			HH_8.clear();
        			HH_9.clear();
        			HH_0.clear();
				Ecalf10hits.clear();
				Ecall20hits.clear();

        			TVector3 HitPos;
        			TVector3 GraPos;

        			
                		HelixClass * currHelix = new HelixClass();
                		currHelix->Initialize_Canonical(a_trk->getPhi(), a_trk -> getD0(), a_trk -> getZ0(), a_trk -> getOmega(), a_trk->getTanLambda(), 3.5);
                		float BushDist[3] = {0, 0, 0};
                		float BushTime = 0;
                		
                        	for(unsigned int j1 = 0; j1 < currClu->getCalorimeterHits().size(); j1++)
                        	{
                                	CalorimeterHit * a_hit = currClu->getCalorimeterHits()[j1];

                                	BushTime = currHelix->getDistanceToPoint((float*)a_hit->getPosition(), BushDist);

                                	if(BushTime > 0)
                                	{
                                	        if(BushDist[2] > _MaxDisHel)
                                        	{
                                                	_MaxDisHel = BushDist[2];
                                        	}
                                        	if(BushDist[2] < _MinDisHel)
                                        	{
                                            	        _MinDisHel = BushDist[2];
                                        	}
                                	}
                        	}
                
        
//////////////////////////////////////
        			

                		

                		int currCluNHits = currClu->getCalorimeterHits().size();
                                if (currCluNHits == 0) continue;
                		_CluNHit += currCluNHits;
                		GraPos = ClusterCoG(currClu);
                   		_graDepth = DisSeedSurface(GraPos);
                		int index1 = 0, index2 = 0;
                		/////////////////////
                		for(int s1 = 0; s1 < currCluNHits; s1++)
                		{
                        		CalorimeterHit * a_hit = currClu->getCalorimeterHits()[s1];
                        		allhits.push_back(a_hit);
                        		int NLayer = idDecoder(a_hit)["K-1"];

                        		HitPos = a_hit->getPosition();
                        		currDepth = DisSeedSurface(HitPos);
                                        float crdis = (CluPos-HitPos).Mag()*sin((CluPos-HitPos).Angle(IntDir));
                        		
                                        if(currDepth > _maxDepth)
                        		{
                                		_maxDepth = currDepth;
                                		index1 = s1;
                        		}
                        		if(currDepth < _minDepth)
                        		{		
                                		_minDepth = currDepth;
                                		index2 = s1;
                        		}

                        		if( fabs(a_hit->getEnergy() - DHCALCalibrationConstant) < 1.0E-6 )      //or other fancy judgements...
                        		{
                                		_HcalNHit++;
                                		_HcalEn += a_hit->getEnergy();
                               		 	Hcalhits.push_back(a_hit);
                                		if(NLayer < 5)
                                		{
                                        		HH_1.push_back(a_hit);
						}
                                		else if(NLayer < 10)
                                		{
                                        		HH_2.push_back(a_hit);
                                		}
                                		else if(NLayer < 15)
                                		{
                                        		HH_3.push_back(a_hit);
                                		}
                                		else if(NLayer < 20)
                                		{
                                        		HH_4.push_back(a_hit);
                                		}
                                		else if(NLayer < 25)
                                		{
                                        		HH_5.push_back(a_hit);
                                		}
                                		else if(NLayer < 30)
                                		{
                                        		HH_6.push_back(a_hit);
                                		}
                                		else if(NLayer < 35)
                                		{
                                        		HH_7.push_back(a_hit);
                                		}
                                		else if(NLayer < 40)
                                		{
                                        		HH_8.push_back(a_hit);
                                		}
                                		else if(NLayer < 45)
                                		{
                                        		HH_9.push_back(a_hit);
                                		}
                                		else
                                		{
                                        		HH_0.push_back(a_hit);
                                		}
                        		}
                        		else
                        		{
                                		_EcalNHit++;
                                		_EcalEn += a_hit->getEnergy();
                                		Ecalhits.push_back(a_hit);
						if(NLayer< 10) Ecalf10hits.push_back(a_hit);
						else Ecall20hits.push_back(a_hit);
                                                if(crdis < 22) _EEClu_R += a_hit->getEnergy();
                                                if(crdis < 11) _EEClu_r += a_hit->getEnergy();
                                                if(crdis < 6) _EEClu_p += a_hit->getEnergy();
                                		if(NLayer < 5)
                                		{
                                        		EH_1.push_back(a_hit);
                                                        _EEClu_L10 += a_hit->getEnergy();
                                		}
                                		else if(NLayer < 10)
                                		{
                                        		EH_2.push_back(a_hit);
                                                        _EEClu_L10 += a_hit->getEnergy();
                                		}
                                		else if(NLayer < 15)
                                		{
                                        		EH_3.push_back(a_hit);
                                		}
                                		else if(NLayer < 20)
                                		{
                                        		EH_4.push_back(a_hit);
                                		}
                                		else if(NLayer < 25)
                                		{
                                        		EH_5.push_back(a_hit);
                                		}
                                		else
                                		{
                                        		EH_6.push_back(a_hit);
                                		}
                         		}
               			}////////////////
               			CalorimeterHit * maxdis_hit = currClu->getCalorimeterHits()[index1];
               			CalorimeterHit * mindis_hit = currClu->getCalorimeterHits()[index2];
               			TVector3 maxpos = maxdis_hit->getPosition();
               			TVector3 minpos = mindis_hit->getPosition();
               			_cluDepth = (maxpos-minpos).Mag();
               			_graAbsDepth = (GraPos-minpos).Mag();
       				_maxDisHtoL = -100;
               			_minDisHtoL = 1E6;
                		_nPhiBump = PhiBumpTag(currClu);
                		float totDisHtoL = 0;
                		float totHitEn = 0;
                		float totHitEnDis = 0;
                		float HitEn;
                
                		for(int s2 = 0; s2 < currCluNHits; s2++)
                		{
                        		CalorimeterHit * a_hit2 = currClu->getCalorimeterHits()[s2];

                        		HitPos = a_hit2->getPosition();
                        		HitEn  = a_hit2->getEnergy();
                        		TVector3 par1 = GraPos-minpos;
                        		TVector3 par2 = minpos-HitPos;
                        		TVector3 par3 = par1.Cross(par2);
                        		float disHtoL = par3.Mag()/par1.Mag();
                        		totDisHtoL+=disHtoL;
                        		totHitEn+=HitEn;
                        		totHitEnDis+=HitEn*disHtoL;
                        		if (disHtoL > _maxDisHtoL) _maxDisHtoL = disHtoL;
                        		if (disHtoL < _minDisHtoL) _minDisHtoL = disHtoL;    

                		}
                		_avDisHtoL = totDisHtoL/currCluNHits;
                		_avEnDisHtoL = totHitEnDis/totHitEn;
             
                 
/////////////////////////
        			_FD_all = FDV2(allhits, ECALCellIDDecoder);
        			_FD_ECAL = FDV2(Ecalhits, ECALCellIDDecoder);
        			_FD_HCAL = FDV2(Hcalhits, ECALCellIDDecoder);
				_FD_ECALF10 = FDV2(Ecalf10hits, ECALCellIDDecoder);
				_NH_ECALF10 = Ecalf10hits.size();
				_FD_ECALL20 = FDV2(Ecall20hits, ECALCellIDDecoder);
                                _NH_ECALL20 = Ecall20hits.size();

        			_NLEcal = 0;
        			_NLHcal = 0;
        			
        			_CluFlagID = 211;

        			for(int p0 = 0; p0 < 8; p0++)
        			{
                			_NH[p0] = 0;
                			_NL[p0] = 0;
                			_FD[p0] = 0;
        			}

        			_FD[0] = FDV2(EH_1, ECALCellIDDecoder);
        			_FD[1] = FDV2(EH_2, ECALCellIDDecoder);
        			_FD[2] = FDV2(EH_3, ECALCellIDDecoder);
        			_FD[3] = FDV2(EH_4, ECALCellIDDecoder);
        			_FD[4] = FDV2(EH_5, ECALCellIDDecoder);
        			_FD[5] = FDV2(EH_6, ECALCellIDDecoder);
        			_FD[6] = FDV2(HH_1, ECALCellIDDecoder);
        			_FD[7] = FDV2(HH_2, ECALCellIDDecoder);
        			_FD[8] = FDV2(HH_3, ECALCellIDDecoder);
        			_FD[9] = FDV2(HH_4, ECALCellIDDecoder);
        			_FD[10] = FDV2(HH_5, ECALCellIDDecoder);
        			_FD[11] = FDV2(HH_6, ECALCellIDDecoder);
        			_FD[12] = FDV2(HH_7, ECALCellIDDecoder);
        			_FD[13] = FDV2(HH_8, ECALCellIDDecoder);
        			_FD[14] = FDV2(HH_9, ECALCellIDDecoder);
        			_FD[15] = FDV2(HH_0, ECALCellIDDecoder);

        			_NH[0] = EH_1.size();
        			_NH[1] = EH_2.size();
        			_NH[2] = EH_3.size();
        			_NH[3] = EH_4.size();
        			_NH[4] = EH_5.size();
        			_NH[5] = EH_6.size();
        			_NH[6] = HH_1.size();
        			_NH[7] = HH_2.size();
        			_NH[8] = HH_3.size();
        			_NH[9] = HH_4.size();
        			_NH[10] = HH_5.size();
        			_NH[11] = HH_6.size();
        			_NH[12] = HH_7.size();
        			_NH[13] = HH_8.size();
        			_NH[14] = HH_9.size();
        			_NH[15] = HH_0.size();

				float sum_NHE = 0, sum_NHH = 0;
				_av_NHE = 0;
   				_av_NHH = 0;
				_AL_Ecal = 0;
				_AL_Hcal = 0;
                                
				for(int r1 = 0; r1 < 16; r1++)
				{
					if(r1 < 6 && _NH[r1]>0)
					{
						sum_NHE += _NH[r1];
						_AL_Ecal++;
					}
					if(r1 >= 6 && _NH[r1]>0)
					{
						sum_NHH += _NH[r1];
						_AL_Hcal++;
					}
				}
                                if(_AL_Ecal > 0)
				        _av_NHE = sum_NHE/_AL_Ecal;
				if(_AL_Hcal > 0)
					_av_NHH = sum_NHH/_AL_Hcal;
				
                                _rms_Ecal = 0;
                                _rms_Hcal = 0;
                                for(int r0 = 0; r0 < 16; r0++)
				{
 					if(r0 < 6)
					{
						if(_NH[r0] > 0)
						{ 
							_rms_Ecal+=pow(_NH[r0]-_av_NHE,2);
							_rms_Ecal2 += pow((double)_NH[r0],2);
						}
					}
					else
					{
						if(_NH[r0] > 0)
						{ 
							_rms_Hcal+=pow(_NH[r0]-_av_NHH,2);
							_rms_Hcal2 += pow((double)_NH[r0],2);
						}							
					} 
				}
				if(_AL_Ecal > 0)
				{
					_rms_Ecal2 = sqrt(_rms_Ecal2/_AL_Ecal); 
					_rms_Ecal = sqrt(_rms_Ecal/_AL_Ecal);
				}
				else
				{
					_rms_Ecal2 = -1; 
					_rms_Ecal = -1;
				}
				if(_AL_Hcal > 0)
				{
					_rms_Hcal2 = sqrt(_rms_Hcal2/_AL_Ecal); 
					_rms_Hcal = sqrt(_rms_Hcal/_AL_Ecal);
				}
				else
				{
					_rms_Hcal2 = -1; 
					_rms_Hcal = -1;
				}
       
        			_NLEcal = ActiveLayers(Ecalhits, ECALCellIDDecoder);
        			_NLHcal = ActiveLayers(Hcalhits, ECALCellIDDecoder);
        			_NL[0] = ActiveLayers(EH_1, ECALCellIDDecoder);
        			_NL[1] = ActiveLayers(EH_2, ECALCellIDDecoder);
        			_NL[2] = ActiveLayers(EH_3, ECALCellIDDecoder);
        			_NL[3] = ActiveLayers(EH_4, ECALCellIDDecoder);
        			_NL[4] = ActiveLayers(EH_5, ECALCellIDDecoder);
        			_NL[5] = ActiveLayers(EH_6, ECALCellIDDecoder);
        			_NL[6] = ActiveLayers(HH_1, ECALCellIDDecoder);
        			_NL[7] = ActiveLayers(HH_2, ECALCellIDDecoder);
        			_NL[8] = ActiveLayers(HH_3, ECALCellIDDecoder);
        			_NL[9] = ActiveLayers(HH_4, ECALCellIDDecoder);
        			_NL[10] = ActiveLayers(HH_5, ECALCellIDDecoder);
        			_NL[11] = ActiveLayers(HH_6, ECALCellIDDecoder);
        			_NL[12] = ActiveLayers(HH_7, ECALCellIDDecoder);
        			_NL[13] = ActiveLayers(HH_8, ECALCellIDDecoder);
        			_NL[14] = ActiveLayers(HH_9, ECALCellIDDecoder);
        			_NL[15] = ActiveLayers(HH_0, ECALCellIDDecoder);

        			

                                bool cutmu1 = _FD_ECAL >= 0 && _FD_HCAL >= 0;
				bool cutmu2;

				if(_cluDepth < 400) cutmu2 = 0;
				else if(_cluDepth < 1200) cutmu2 = _FD_all < 0.1/pow(500.0,2)*pow((_cluDepth-800),2)+0.05;
				else cutmu2 = 1;

				bool cutmu3;
				if(_TrkEn > 70) cutmu3 = _FD_all < (600./_avEnDisHtoL + 20)/100.;
				else if (_TrkEn > 10) cutmu3 = _FD_all < (600./_avEnDisHtoL -10 + 0.5*_TrkEn)/100.;
				else if (_TrkEn > 7.5) cutmu3 = _FD_all < (600./_avEnDisHtoL -10)/100.;
				else cutmu3 = 1;
                                bool cutmu3b;
				if (_TrkEn > 10) cutmu3b = _avEnDisHtoL < 25;
				else cutmu3b = _avEnDisHtoL < 25 + 10 * (10 - _TrkEn);
				bool cutmu8en = cutmu1 && cutmu2 && cutmu3 && cutmu3b;
                                
				bool cutmu4 = _FD_HCAL < 0;
				bool cutmu5 = _FD_ECAL < 0.1/10.*_av_NHE-0.05 &&_av_NHE < 15; 
				bool cutmu6 = _cluDepth > 1200;
				bool cutmu7 = _FD_all < 0.3/sqrt(400.)*sqrt(_cluDepth-1200.);
				bool cutmu8 = _rms_Hcal < 10 && _FD_HCAL < -0.25/600.*_MaxDisHel+0.25;
				bool cutmu9 = _cluDepth > 600 && _EEClu_p/_EcalEn < 0.2;

				bool cutmu;
				if(_TrkEn > 8) cutmu = cutmu8en;
				else
				{
					if(cutmu4)
					{
						if(cutmu5) cutmu = 1;
						else cutmu = 0;
					}
					else
					{
						if(_TrkEn < 2.5)
						{
							if(cutmu9) cutmu = 1;
							else cutmu = 0;
						}
						else if(_TrkEn < 4)
						{
							if(cutmu6 && cutmu7) cutmu = 1;
							else cutmu = 0;
						}
						else
						{
							if(cutmu6 && cutmu7 && cutmu8) cutmu = 1;
							else cutmu = 0;
						}
					}
				}

                                bool cute1 = _FD_HCAL == -1;
				bool cute2 = _EcalEn/(_EcalEn+_HcalEn) > 0.8;
				bool cute3 = _FD_all > 0.1;
				bool cute4 = log10(_EEClu_r) > 0.01*(_cluDepth-400);
				bool cute5 = _avEnDisHtoL < 15;
				bool cute6 = _FD_ECAL >= 0.6/1.5*(log10(_EcalNHit)-1);
				bool cute7 = _FD_all >= 0.6/1.5*(log10(_EcalNHit+2.*_HcalNHit)-1.2);
                                bool cute8 = _cluDepth < 400+20/3.*(_TrkEn-10);
				bool cute9 = _FD_ECALF10 > 0.2;
				bool cute10 = _FD_ECALF10 > 0.5/400.*_cluDepth;
				bool cute11 = _FD_ECALL20 != 0;
                                bool cute;

				if(_TrkEn < 6) cute = cute1 && cute2 && cute3 && cute4 && cute5 && cute6 && cute8 && cute9;
				else if (_TrkEn < 15) cute = cute2 && cute3 && cute4 && cute6 && cute8 && cute9 && cute10 && cute11;
				else cute = cute2 && cute3 && cute4 && cute7 && cute8 && cute9 && cute10; 
                                
                                if(cutmu) _CluFlagID = 13;
                                else if (cute)  _CluFlagID = 11;

                                if (_CluFlagID == 13) _nMuPFO++;
				else if (_CluFlagID == 11) _nEPFO++;
                                else _nPiPFO++; 

                                delete currHelix;
                                 
                                _outputPFO->Fill();
                           
                            }
                           





                        }
                    }
                    _nCH = nChPFO;
                    _nMIPs = nMIP;
                    _nEMs = nEM;
                    _nHADs = nHAD;

		    if(_nCH > 0) _thetaDisCh = _thetaMC;
		    else _thetaDisCh = -1;
		    if(_nMIPs > 0) _thetaDisMip = _thetaMC;
		    else _thetaDisMip = -1;
		    if(_nEMs > 0) _thetaDisEm = _thetaMC;
		    else _thetaDisEm = -1;
		    if(_nHADs > 0) _thetaDisHad = _thetaMC;
		    else _thetaDisHad = -1;

		    

                    _outputTree->Fill();

		    num++;
                }catch (lcio::DataNotAvailableException err) { }
         }
}

void Trknum::end()
{

	if (_outputTree) {

		TFile *tree_file = _outputTree->GetCurrentFile(); //just in case we switched to a new file
		tree_file->Write();
		delete tree_file;
	}

}






