#include <BushConnect.hh>
#include <ArborTool.hh>
#include <ArborToolLCIO.hh>
#include <DetectorPos.hh>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/CalorimeterHit.h>
#include <EVENT/SimCalorimeterHit.h>
#include <EVENT/LCFloatVec.h>
#include <EVENT/LCParameters.h>
#include <EVENT/Cluster.h>
#include <EVENT/LCRelation.h>

#include <IMPL/ReconstructedParticleImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCFlagImpl.h>
#include <IMPL/LCRelationImpl.h>
#include <IMPL/ClusterImpl.h>
#include "UTIL/CellIDDecoder.h"

#include "HelixClass.h"		//in Marlin Util

#include <values.h>
#include <string>
#include <iostream>
#include <cmath>
#include <vector>
#include <stdexcept>
#include <cstdlib>
#include <TMath.h>
#include <TFile.h>
#include <TTree.h>
#include <Rtypes.h>
#include <sstream>
#include <TH1.h>
#include <TVector3.h>
#include <TMath.h>

using namespace std;
// using namespace arbor; 

const float TightGeoThreshold = 30;
const string ECALCellIDDecoder  = "M:3,S-1:3,I:9,J:9,K-1:6";

BushConnect aBushConnect ;
BushConnect::BushConnect()
	: Processor("BushConnect"),
	_output(0)
{
	_description = "Track Cluster Linking. Track info represented by MCTruth at this moment" ;
        SCale1 = 0.5;
        registerProcessorParameter( "Scale1" ,
                        "scale1." ,
                         SCale1,
                         SCale1);
}

void BushConnect::init() {

	printParameters();

	Cluflag.setBit(LCIO::CHBIT_LONG);
}

void BushConnect::Clean(){

	Track_Energy.clear();
	Track_Type.clear();
	SortedTracks.clear();

	ClusterType_1stID.clear();
	ChCoreID.clear();

	ecalchcore_tight.clear();
        ecalchcore_medium.clear();
        ecalchcore_loose.clear();
	ecalchcore.clear();
        ecalnecore_EM.clear();
	ecalnecore_NonEM.clear();
        ecalfrag.clear();
        ecalundef.clear();
	ecalfrag_TBM_CH.clear();
	ecalfrag_TBM_NE_EM.clear();
	ecalfrag_TBM_NE_NonEM.clear();
	ecalpotentialbackscattering.clear();
	ecalundef_iso.clear();

	hcalchcore_tight.clear();
	hcalchcore_medium.clear();
	hcalchcore_loose.clear();
	hcalnecore.clear();
	hcalchcore.clear();
	hcalfrag.clear();
	hcalfrag_TBM_CH.clear();
	hcalfrag_TBM_NE_EM.clear();
	hcalfrag_TBM_NE_NonEM.clear();
	hcalundef.clear();
	hcalundef_iso.clear();

	trkendposition.clear();

	chargedclustercore.clear();
	non_chargedclustercore.clear();		//all clusters besides charged cluster core
}

void BushConnect::TrackSort(LCEvent* evtPP) //, &std::map<Track*, int>Track_Tpye, &std::map<Track*, float> Track_Energy)
{

	LCCollection * TrackColl = evtPP->getCollection("MarlinTrkTracks");

	int NTrk = TrackColl->getNumberOfElements();
	float D0 = 0;
	float Z0 = 0;
	int NTrkHit = 0;
	const float mass = 0.139;	//Pion Mass
	TVector3 EndPointPos, StartPointPos; 
	int TrackType = 0; 

	std::vector<Track*> tracks_HQ_Barrel; 
	std::vector<Track*> tracks_HQ_Endcap;
	std::vector<Track*> tracks_HQ_Shoulder;
	std::vector<Track*> tracks_HQ_Forward; 
	std::vector<Track*> tracks_MQ_Barrel;
	std::vector<Track*> tracks_MQ_Endcap;
	std::vector<Track*> tracks_MQ_Shoulder;
	std::vector<Track*> tracks_MQ_Forward;
	std::vector<Track*> tracks_Vtx; 
	std::vector<Track*> tracks_LQ; 
	std::vector<Track*> curr_tracks;

	trkendposition.clear();

	tracks_HQ_Barrel.clear();
	tracks_HQ_Endcap.clear();
	tracks_HQ_Shoulder.clear();
	tracks_HQ_Forward.clear();
	tracks_MQ_Barrel.clear();
	tracks_MQ_Endcap.clear();
	tracks_MQ_Shoulder.clear();
	tracks_MQ_Forward.clear();
	tracks_Vtx.clear();
	tracks_LQ.clear();

        std::vector<Track*> tracks_ILL;
        tracks_ILL.clear();


	std::vector<int> TrackOrder; 
	TrackOrder.clear();	
	std::map<Track*, int> Track_Index; 
	Track_Index.clear();
	Track_Energy.clear();
	Track_Type.clear();

	for(int t0 = 0; t0 < NTrk; t0++)
	{
		Track* a_Trk = dynamic_cast<Track*>( TrackColl->getElementAt( t0 ) );
		D0 = a_Trk->getD0();
		Z0 = a_Trk->getZ0();
		NTrkHit = a_Trk->getTrackerHits().size();
		TrackerHit * last_hit = a_Trk->getTrackerHits()[NTrkHit - 1];
		EndPointPos = last_hit->getPosition();
		trkendposition[a_Trk] = EndPointPos;
                StartPointPos = (a_Trk->getTrackerHits()[0])->getPosition();
		Track_Index[a_Trk] = t0;

		if( NTrkHit > 9 || (fabs(EndPointPos.Z()) > LStar - 500 && EndPointPos.Perp() < TPCInnerRadius ) || fabs(EndPointPos.Z()) > ECALHalfZ - 200  )		// Min requirement for track quality
		{	// LStar - 500, suppose to be the last Disk Position

			TrackType = 0;

			if( fabs(EndPointPos.Z()) > ECALHalfZ - 500 && EndPointPos.Perp() > TPCOuterRadius - 300  )	//Shoulder
			{
				TrackType = 30;
			}
			else if( fabs(EndPointPos.Z()) > LStar - 500 && EndPointPos.Perp() < TPCInnerRadius )		//Forward
			{
				TrackType = 40;
			}
			else if( EndPointPos.Perp() > TPCOuterRadius - 100 )		//Barrel
			{
				TrackType = 10;
			}
			else if( fabs(EndPointPos.Z()) > ECALHalfZ - 200 )		//Endcap
			{
				TrackType = 20; 
			}

			if( fabs(D0) < 1 && fabs(Z0) < 1 )
			{
				TrackType += 1;
			}

			Track_Type[a_Trk] = TrackType;
			
			if(TrackType == 41)
                                tracks_HQ_Forward.push_back(a_Trk);
			else if(TrackType == 40)
                                tracks_MQ_Forward.push_back(a_Trk); 
			//else if( StartPointPos.Mag() > 300 && NTrkHit < 50  )
                        //        tracks_ILL.push_back(a_Trk);
			else if(TrackType == 21)
                                tracks_HQ_Endcap.push_back(a_Trk);
			else if(TrackType == 20)
                                tracks_MQ_Endcap.push_back(a_Trk);
			else if(TrackType == 11)
				tracks_HQ_Barrel.push_back(a_Trk);
			else if(TrackType == 31)
				tracks_HQ_Shoulder.push_back(a_Trk);
			else if(TrackType == 10)
				tracks_MQ_Barrel.push_back(a_Trk);
			else if(TrackType == 30)
				tracks_MQ_Shoulder.push_back(a_Trk);
			else if(TrackType == 1)
				tracks_Vtx.push_back(a_Trk);
                        //else if( StartPointPos.Mag() > 50 && EndPointPos.Mag() < 1000 && NTrkHit < 50  )
                          //      tracks_ILL.push_back(a_Trk);
			else
				tracks_LQ.push_back(a_Trk);
		}
	}

	std::vector<float > currTrkMomentum;
	std::vector<int> currTrkIndex;

	for(int t1 = 0; t1 < 10; t1++)
	{
		currTrkMomentum.clear();
		currTrkIndex.clear();
		curr_tracks.clear();
		if(t1 == 0)
			curr_tracks = tracks_HQ_Endcap;
		else if(t1 == 1)
			curr_tracks = tracks_HQ_Barrel;
		else if(t1 == 2)
			curr_tracks = tracks_MQ_Endcap;
		else if(t1 == 3)
			curr_tracks = tracks_MQ_Barrel;
		else if(t1 == 4)
			curr_tracks = tracks_HQ_Shoulder;
		else if(t1 == 5)
			curr_tracks = tracks_MQ_Shoulder;
		else if(t1 == 6)
			curr_tracks = tracks_HQ_Forward;
		else if(t1 == 7)
			curr_tracks = tracks_MQ_Forward;
		else if(t1 == 8)
			curr_tracks = tracks_Vtx;
		else if(t1 == 9)			
			curr_tracks = tracks_LQ; 

		int N_currTrack = curr_tracks.size();
		float TrackEn = 0; 

		for(int t2 = 0; t2 < N_currTrack; t2++)
		{
			Track* tmpTrk = curr_tracks[t2];
			HelixClass * TrkInit_Helix = new HelixClass();
			TrkInit_Helix->Initialize_Canonical(tmpTrk->getPhi(), tmpTrk -> getD0(), tmpTrk -> getZ0(), tmpTrk -> getOmega(), tmpTrk->getTanLambda(), BField);
			TrackEn = mass*mass;

			for (int q3 = 0; q3 < 3; q3 ++)
			{
				TrackEn += (TrkInit_Helix->getMomentum()[q3])*(TrkInit_Helix->getMomentum()[q3]);
			}

			TrackEn = sqrt(TrackEn);
			currTrkMomentum.push_back(TrackEn);
			Track_Energy[tmpTrk] = TrackEn; 		

			delete TrkInit_Helix;
		}

		currTrkIndex = SortMeasure(currTrkMomentum, 1);

		for(int t3 = 0; t3 < N_currTrack; t3++)
		{
			Track* b_tmpTrk = curr_tracks[currTrkIndex[t3]];
			// cout<<t1<<"th collection: currTrackEnergy: "<<Track_Energy[b_tmpTrk]<<endl; 
			if(t1 < 9 || Track_Energy[b_tmpTrk] < 10)
				TrackOrder.push_back(Track_Index[b_tmpTrk]);
		}
	}

	for(unsigned int t4 = 0; t4 < TrackOrder.size(); t4++)
	{
		Track* b_Trk = dynamic_cast<Track*>( TrackColl->getElementAt( TrackOrder[t4] ) );
		SortedTracks.push_back(b_Trk);
		// cout<<"Track Order... "<<Track_Type[b_Trk]<<" : "<<Track_Energy[b_Trk]<<endl;
	}

	// cout<<"NTrk Size:  "<<NTrk<<" vs "<<SortedTracks.size()<<endl; 

}

void CheckConsistence(LCEvent *evtQQ, std::vector<std::string> SubClass)
{
	float inputEn = 0;
	int inputNCl = 0;
	float TotalEn = 0;
	int TotalNCl = 0;

	for(int c1 = 0; c1 < SubClass.size(); c1++)
	{

		LCCollection * LocalCol = evtQQ->getCollection( SubClass[c1].c_str() );

		int CurrCollSize = LocalCol->getNumberOfElements();
		float CurrEnergy = 0;

		for(int c2 = 0; c2 < CurrCollSize; c2++)
		{
			Cluster * a_clu = dynamic_cast<Cluster*>(LocalCol->getElementAt(c2));
			CurrEnergy += a_clu->getEnergy();
		}

		if(c1 < 2)
		{
			inputEn += CurrEnergy;
			inputNCl += CurrCollSize;
		}
		else
		{
			TotalNCl += CurrCollSize;
			TotalEn += CurrEnergy;
		}

		cout<<"CurrColl "<<SubClass[c1].c_str()<<" has "<<CurrCollSize<<" Clusters with total En = "<<CurrEnergy<<" GeV"<<endl;
	}

	cout<<endl<<"Comparison Energy "<<inputEn<<" =? "<<TotalEn<<endl;
	cout<<"Comparison Cluster Size "<<inputNCl<<" =? "<<TotalNCl<<endl<<endl;
}

void BushConnect::TagCore(LCEvent * evtPP) //
{
	//Let's define several distance Matrix
	//	Track - ECAL Clusters
	//	Type 1 ECAL Cluster: Close to Track extrapolation
	//	Type 2 ECAL Cluster: 

	LCCollection * ECALClu = evtPP->getCollection("EcalBushes");	//A sort here should be helpful
	LCCollection * HCALClu = evtPP->getCollection("HcalBushes");
	int NClu_Hcal = HCALClu->getNumberOfElements();
	int NClu_Ecal = ECALClu->getNumberOfElements();
	int NTrk = SortedTracks.size();
	std::vector<Cluster* > ecalbushes = CollClusterVec(ECALClu);
	std::vector<Cluster* > hcalbushes = CollClusterVec(HCALClu);
	TVector3 CluPos;

	/*

	std::vector<float> _usable_ecal_clusters_En;
	std::vector<float> _usable_hcal_clusters_En; 
	_usable_ecal_clusters_En.clear();
	_usable_hcal_clusters_En.clear();
	_usable_ecal_clusters_En.reserve(NClu_Ecal);
        _usable_ecal_clusters_En.reserve(NClu_Hcal);
	
	std::vector<int> _usable_ecal_clusters;
        std::vector<int> _usable_hcal_clusters;
        _usable_ecal_clusters.clear();
        _usable_hcal_clusters.clear();
        _usable_ecal_clusters.reserve(NClu_Ecal);
        _usable_ecal_clusters.reserve(NClu_Hcal);

        for(int P = 0; P < NClu_Ecal; P++)
        {
		_usable_ecal_clusters_En.push_back( ecalbushes[P]->getEnergy() );
        }
	
	for(int P = 0; P < NClu_Hcal; P++)
        {
                _usable_hcal_clusters_En.push_back( hcalbushes[P]->getEnergy() );
        }

	_usable_ecal_clusters = SortMeasure(_usable_ecal_clusters_En, 1);
	_usable_hcal_clusters = SortMeasure(_usable_hcal_clusters_En, 1);

	for(int i = 0; i < NClu_Ecal; i++)
	{
		cout<<i<<" : "<<ecalbushes[ _usable_ecal_clusters[i] ]->getEnergy()<<endl;
	}
	
	*/
	std::map<Cluster*, int> BushTouchFlag; 
	std::map<Track*, int> TrkTouchFlag; 
	std::map<Track*, Cluster*> FCMap_Track_ABSCluster; 	//Careful about 1 - 1 on to...
	std::map<Track*, float> Map_Track_ClusterEnergy; 
	std::map<Track*, Cluster*> FCMap_Track_EcalCore;
	std::map<Track*, Cluster*> FCMap_Track_HcalCore;
	std::map<Track*, Cluster*> FCMap_Track_CHCore;
	BushTouchFlag.clear();
	TrkTouchFlag.clear();
	FCMap_Track_ABSCluster.clear();
	FCMap_Track_EcalCore.clear();
	FCMap_Track_HcalCore.clear();
	FCMap_Track_CHCore.clear();
	Map_Track_ClusterEnergy.clear();

	float currTrkEn = 0; 
	int currTrackType = 0;
	float MinimalTrkBushDis_E = 1.0E9;
	float MinimalTrkBushDis_H = 1.0E9;
	float DisMatrix_Track_Clu_E[NTrk][NClu_Ecal];
	// float DisMatrix_Track_Clu_H[NTrk][NClu_Hcal];
	int CluIndex = -1; 
	int CluIndex_H = -1; 
	int ClosestCluType = 0; 
	float CluDepth = 0; 

	//Debug
	/*
	   float ChCoreEn = 0;
	   float NeCoreEn = 0;
	   float FragEn = 0;
	   float UndefEn = 0;
	   float UndefEn1 = 0; 
	   float UndefEn2 = 0; 
	   float TBMEn = 0; 
	//E Debug
	 */

	LCCollection *checalcorecluster = new LCCollectionVec(LCIO::CLUSTER);
	LCCollection *chhcalcorecluster = new LCCollectionVec(LCIO::CLUSTER);
	LCCollection *chcorecluster = new LCCollectionVec(LCIO::CLUSTER);
	checalcorecluster->setFlag(Cluflag.getFlag());
	chhcalcorecluster->setFlag(Cluflag.getFlag());
	chcorecluster->setFlag(Cluflag.getFlag());

	for(int s0 = 0; s0 < NTrk; s0++)
	{
		for(int s1 = 0; s1 < NClu_Ecal; s1++)
		{
			DisMatrix_Track_Clu_E[s0][s1] = 1.0E10; 
		}
		/*
		   for(int s2 = 0; s2 < NClu_Hcal; s2++)
		   {
		   DisMatrix_Track_Clu_H[s0][s2] = 1.0E10; 
		   }
		 */
	}

	std::vector<Cluster*> TightLinkedCluster; 
	std::vector<Cluster*> TightLinkedCluster_Hcal; 
	std::vector<Cluster*> TightLinkedCluster_Ecal;

	float EcalCoreEnergy = 0; 
	float HcalCoreEnergy = 0;
	float CoreMergeDistanceDepthCorrector = 0; 

	TVector3 TrkEndPoint;

	for(int i0 = 0; i0 < NTrk; i0++)  //Dropped Size can exist
	{
		Track* a_trk = SortedTracks[i0];
		currTrackType = Track_Type[a_trk];
		currTrkEn = Track_Energy[a_trk];

		EcalCoreEnergy = 0;
		HcalCoreEnergy = 0;

		MinimalTrkBushDis_E = 1.0E9;
		MinimalTrkBushDis_H = 1.0E9;
		CluIndex = -1;
		CluIndex_H = -1;  

		TrkEndPoint = trkendposition[a_trk];

		TightLinkedCluster.clear();
		TightLinkedCluster_Ecal.clear();
		TightLinkedCluster_Hcal.clear();

		// TrackHit * endhit = a_trk->getTrackerHits()[]

		for(int j0 = 0; j0 < NClu_Ecal; j0++)
		{
			Cluster *fccand_bush = ecalbushes[j0];			
			float Dis = SimpleDisTrackClu(a_trk, fccand_bush);
			CluPos = fccand_bush->getPosition();
			CluDepth = DisSeedSurface(CluPos);
			CoreMergeDistanceDepthCorrector = 0;
			if(CluDepth > 20)
				CoreMergeDistanceDepthCorrector = 20;
			else if(CluDepth > 10)
				CoreMergeDistanceDepthCorrector = 10;
			else if(CluDepth > 5)
				CoreMergeDistanceDepthCorrector = 5; 

			float TrackEndPDis = (TrkEndPoint - CluPos).Mag();

			if(Dis > -0.1)
			{
				DisMatrix_Track_Clu_E[i0][j0] = Dis;
			}
			if(Dis < MinimalTrkBushDis_E && Dis > -0.1 && fccand_bush->getCalorimeterHits().size() > 3 && (TrackEndPDis < 400 || currTrackType != 11) )
			{
				MinimalTrkBushDis_E = Dis;
				CluIndex = j0;
			}
			//if(Dis < 7 + CoreMergeDistanceDepthCorrector && Dis > -0.1 && BushTouchFlag.find(fccand_bush) == BushTouchFlag.end() && (currTrkEn > 3 || fccand_bush->getEnergy() < 5 ) && (TrackEndPDis < 400 || currTrackType != 11))
			if(Dis < 7 + CoreMergeDistanceDepthCorrector && Dis > -0.1 && BushTouchFlag.find(fccand_bush) == BushTouchFlag.end() && (currTrkEn > 3 || fccand_bush->getEnergy() < 5 ) && (TrackEndPDis < 400 || currTrackType != 11))
			{
				/*
				   if(currTrackType == 11)
				   cout<<fccand_bush->getEnergy()<<" : "<< TightLinkedCluster.size()<<" : "<<CluPos.X()<<" : "<<CluPos.Y()<<" : "<<CluPos.Z()<<" -----  "<<TrkEndPoint.X()<<" : "<<TrkEndPoint.Y()<<" : "<<TrkEndPoint.Z()<<" ~~ Distance ~~"<<TrackEndPDis<<endl;
				 */
				TightLinkedCluster.push_back(fccand_bush);
				TightLinkedCluster_Ecal.push_back(fccand_bush);
				BushTouchFlag[fccand_bush] = currTrackType;
			}
		}

		if(CluIndex > -1)
		{

			Cluster * a_close_tree = ecalbushes[CluIndex];
			ClosestCluType = 0;

			if( MinimalTrkBushDis_E< 10.0 && currTrkEn > 1.0 )    // Maybe some restrictions on depth? Needed for ECAL?? 
			{    									//Efficiency & Purity...
				ClosestCluType = 1; 
			}
			else if( MinimalTrkBushDis_E < 30.0 && currTrackType > 10)
			{
				ClosestCluType = 2;
			}
			else if( MinimalTrkBushDis_E < 150 && MinimalTrkBushDis_E > 30.0 && currTrackType < 5)	//1 or 0, low qualities Big Distances
			{
				ClosestCluType = 3;
			}

			if(ClosestCluType > 0 && BushTouchFlag.find(a_close_tree) == BushTouchFlag.end() )
			{
				if(ClosestCluType == 1)
					ecalchcore_tight.push_back(a_close_tree);
				else if(ClosestCluType == 2)
					ecalchcore_medium.push_back(a_close_tree);
				else if(ClosestCluType == 3)
					ecalchcore_loose.push_back(a_close_tree);

				BushTouchFlag[a_close_tree] = currTrackType;
				ecalchcore.push_back(a_close_tree);
				// ChCoreEn += a_close_tree->getEnergy();
				TightLinkedCluster.push_back(a_close_tree);
				TightLinkedCluster_Ecal.push_back(a_close_tree);
			}
		}

		if( TightLinkedCluster_Ecal.size() > 0 )
		{
			ClusterImpl * chcorecluster_ecal =  NaiveMergeClu(TightLinkedCluster_Ecal);
			checalcorecluster->addElement(chcorecluster_ecal);
			EcalCoreEnergy = chcorecluster_ecal->getEnergy();
			FCMap_Track_EcalCore[a_trk] = chcorecluster_ecal;
		}

		if(currTrkEn > 1.0)	// 
		{
			for(int j1 = 0; j1 < NClu_Hcal; j1++)
			{
				Cluster *fccand_bush_h = hcalbushes[j1];
				float DisH = SimpleDisTrackClu(a_trk, fccand_bush_h);
				/*
				   if(DisH > -0.1)
				   {	
				   DisMatrix_Track_Clu_H[i0][j1] = DisH;
				   }
				 */
				CluPos = fccand_bush_h->getPosition();
				CluDepth = DisSeedSurface(CluPos);

				float TrackEndPDis = (TrkEndPoint - CluPos).Mag();

				if(DisH < MinimalTrkBushDis_H && DisH > 0 && fccand_bush_h->getCalorimeterHits().size() > 3)
				{
					MinimalTrkBushDis_H = DisH;
					CluIndex_H = j1;
				}
				if(DisH < 15 + CluDepth*0.02 && DisH > -0.1 && BushTouchFlag.find(fccand_bush_h) == BushTouchFlag.end() && ( TrackEndPDis < 300 + CluDepth || currTrackType != 11 ) ) // + 0.02*CluDepth)  
				{
					TightLinkedCluster.push_back(fccand_bush_h);
					TightLinkedCluster_Hcal.push_back(fccand_bush_h);
					BushTouchFlag[fccand_bush_h] = currTrackType;
					// cout<<i0<<" : "<<j1<<" Inside Loop "<<DisH<<" - - - "<<TightLinkedCluster_Hcal.size()<<endl;
				}
			}

			if(CluIndex_H > -1)
			{
				Cluster *a_close_tree = hcalbushes[CluIndex_H];
				if(BushTouchFlag.find(a_close_tree) == BushTouchFlag.end() && MinimalTrkBushDis_H < 40.0 )
				{
					BushTouchFlag[a_close_tree] = currTrackType;
					TightLinkedCluster_Hcal.push_back(a_close_tree);
					TightLinkedCluster.push_back(a_close_tree);
				}
			}

			if( TightLinkedCluster_Hcal.size() > 0 )
			{
				ClusterImpl * chcorecluster_hcal =  NaiveMergeClu(TightLinkedCluster_Hcal);
				chhcalcorecluster->addElement(chcorecluster_hcal);
				HcalCoreEnergy = chcorecluster_hcal->getEnergy();
				FCMap_Track_HcalCore[a_trk] = chcorecluster_hcal;
			}
		}

		if( TightLinkedCluster.size() > 0 ) // && EcalCoreEnergy + HcalCoreEnergy < 2.0*currTrkEn )...
		{
			ClusterImpl * chcorecluster_eh =  NaiveMergeClu(TightLinkedCluster);
			chcorecluster->addElement(chcorecluster_eh);
			FCMap_Track_CHCore[a_trk] = chcorecluster_eh;
			Map_Track_ClusterEnergy[a_trk] = chcorecluster_eh->getEnergy();
			chargedclustercore.push_back(chcorecluster_eh);
		}

		cout<<"TrkEn/CluEn "<<currTrackType<<" : "<<currTrkEn<<" / "<<EcalCoreEnergy + HcalCoreEnergy<<" TightLinkedClusterSize "<<TightLinkedCluster_Ecal.size()<<" HCALTightSize "<<TightLinkedCluster_Hcal.size() <<" : "<<TightLinkedCluster.size()<<endl;
	}

	evtPP->addCollection(checalcorecluster, "ECALCORE_CH");
	evtPP->addCollection(chhcalcorecluster, "HCALCORE_CH");
	evtPP->addCollection(chcorecluster, "CHARGEDCORE_BeforeABS");

	// ID Neutral Core
	// ID To be Merged

	float DisToClosestTrack = 1.0E10;
	int CluSize = 0; 
	int ClosetTrackIndex = -1;
	float CluEnergy = 0; 
	float DepthShiftTBM = 0; 
	float DisToCore = 0; 
	int FlagBackScattering = 0; 

	for(int i1 = 0; i1 < NClu_Ecal; i1++)
	{
		Cluster *a_bush = ecalbushes[i1];

		if(BushTouchFlag.find(a_bush) == BushTouchFlag.end())	//Might be a neutral core
		{
			DisToClosestTrack = 1.0E10;
			DisToCore = 1.0E10; 
			CluPos = a_bush->getPosition();
			CluDepth = DisSeedSurface(CluPos);
			CluEnergy = a_bush->getEnergy();
			CluSize = a_bush->getCalorimeterHits().size();
			ClosetTrackIndex = -1;
			FlagBackScattering = 0;	

			for(int j1 = 0; j1 < NTrk; j1++)
			{
				if( DisMatrix_Track_Clu_E[j1][i1] < DisToClosestTrack )
				{
					DisToClosestTrack = DisMatrix_Track_Clu_E[j1][i1];
					ClosetTrackIndex = j1; 
				}
			}

			if(DisToClosestTrack < 100 + CluDepth)
			{
				Cluster *currcore = FCMap_Track_EcalCore[ SortedTracks[ClosetTrackIndex] ];
				if(currcore)
				{
					int Currcoresize = currcore->getCalorimeterHits().size();
					DisToCore = BushDis(currcore, a_bush);
					if(DisToCore < 7 && CluDepth > 5 && Currcoresize < 20)	//Cell Size, close attached
					{
						FlagBackScattering = 1; 	//Create a dedicated collection; then
						ecalpotentialbackscattering.push_back(a_bush);
					}	
				}
			}

			// cout<<"DisToCore "<<DisToClosestTrack<<" : "<<DisToCore<<endl; 

			if(!FlagBackScattering)	//Only Potential
			{

				if(CluEnergy < 0.05 && CluDepth > 30)
				{
					ecalfrag.push_back(a_bush);
				}
				else if( (DisToClosestTrack > 15 && DisToCore > 15) || CluEnergy > 2.0 )     //FD ~ see if ... should enlarge this distances...
				{
					if(CluDepth < 20 && CluEnergy > 0.3) // && FD, MainDirection
					{
						ecalnecore_EM.push_back(a_bush);
					}
					else if(CluEnergy > 0.5)
					{
						ecalnecore_NonEM.push_back(a_bush);
					}
					else
					{
						ecalundef.push_back(a_bush);
					}
				}
				else if(DisToClosestTrack < 60 + CluDepth)	//T.B.M	
				{
					if(CluDepth > 40) 
						DepthShiftTBM = 20;
					else 
						DepthShiftTBM = 0.5*CluDepth;					

					if( DisToCore < 7 + DepthShiftTBM )	//Neighbours?... ~ same level
					{
						ecalfrag_TBM_CH.push_back(a_bush);		//Direct Merging
						// TBMEn += CluEnergy; 
					}
					else	// Potentially loose Core
					{
						ecalundef.push_back(a_bush);
					}
				}
				else
				{
					ecalundef.push_back(a_bush);
					// UndefEn += CluEnergy;
				}
			}		
		}
	}

	int NUndef = ecalundef.size();
	int NNeCoreEM = ecalnecore_EM.size();
	int NNeCoreNonEM = ecalnecore_NonEM.size();
	int NChCore = chcorecluster->getNumberOfElements();
	int NCore = chcorecluster->getNumberOfElements() + NNeCoreEM + NNeCoreNonEM;
	float DisMatrix[NUndef][NCore];

	float tmpDis = 0; 
	float MinDistanceToChCore = 1.0E10; 
	float MinDistanceToNeCore_EM = 1.0E10; 
	float MinDistanceToNeCore_NonEM = 1.0E10; 

	for(int i2 = 0; i2 < NUndef; i2++)
	{
		Cluster *a_Undef = ecalundef[i2];
		MinDistanceToChCore = 1.0E10;
		MinDistanceToNeCore_EM = 1.0E10;
		MinDistanceToNeCore_NonEM = 1.0E10; 

		CluPos = a_Undef->getPosition();
		CluDepth = DisSeedSurface(CluPos);

		for(int j2 = 0; j2 < NChCore; j2++)
		{
			Cluster *a_chcore = dynamic_cast<Cluster*>(chcorecluster->getElementAt(j2));
			tmpDis = BushDis(a_chcore, a_Undef);
			DisMatrix[i2][j2] = BushDis(a_chcore, a_Undef);
			if(tmpDis < MinDistanceToChCore)
			{
				MinDistanceToChCore = tmpDis;
			}
		}
		for(int j3 = 0; j3 < NNeCoreEM; j3++)
		{
			Cluster *a_necore = ecalnecore_EM[j3];
			tmpDis = BushDis(a_necore, a_Undef);
			DisMatrix[i2][j3 + NChCore] = BushDis(a_necore, a_Undef);
			if(tmpDis < MinDistanceToNeCore_EM)
			{
				MinDistanceToNeCore_EM = tmpDis;
			}
		}
		for(int j4 = 0; j4 < NNeCoreNonEM; j4++)
		{
			Cluster *a_necore = ecalnecore_NonEM[j4];
			tmpDis = BushDis(a_necore, a_Undef);
			DisMatrix[i2][j4 + NChCore + NNeCoreEM] = BushDis(a_necore, a_Undef);
			if(tmpDis < MinDistanceToNeCore_NonEM)
			{
				MinDistanceToNeCore_NonEM = tmpDis;
			}
		}

		// cout<<"Min Distance to Ne/Ch for "<<i2<<"th depth & fraction: "<<CluDepth<<" : "<<MinDistanceToChCore<<" : "<<MinDistanceToNeCore_EM<<" : "<<MinDistanceToNeCore_NonEM<<endl;

		if(MinDistanceToChCore < 20 && MinDistanceToChCore < MinDistanceToNeCore_EM && MinDistanceToChCore < MinDistanceToNeCore_NonEM  )
		{
			ecalfrag_TBM_CH.push_back(a_Undef);
		}
		else if(MinDistanceToNeCore_EM < 40 && MinDistanceToNeCore_EM < MinDistanceToChCore && MinDistanceToNeCore_EM < MinDistanceToNeCore_NonEM)	//Distance should be depends on Energy & Depth As well
		{
			ecalfrag_TBM_NE_EM.push_back(a_Undef);
		}
		else if(MinDistanceToNeCore_NonEM < 40 && MinDistanceToNeCore_NonEM < MinDistanceToChCore && MinDistanceToNeCore_NonEM < MinDistanceToNeCore_EM)
		{
			ecalfrag_TBM_NE_NonEM.push_back(a_Undef);
		}
		else
		{
			ecalundef_iso.push_back(a_Undef);
		}
	}

	//HCAL

	std::vector<Cluster*> hcalundef; 
	std::vector<Cluster*> hcalcore; 

	for(int i3 = 0; i3 < NClu_Hcal; i3++)
	{
		Cluster * a_hcalClu = hcalbushes[i3];
		if( BushTouchFlag.find(a_hcalClu) == BushTouchFlag.end() )
		{
			hcalundef.push_back(a_hcalClu);
		}
	}

	LCCollection * hcalundefcoll = ClusterVecColl(hcalundef);

	evtPP->addCollection(hcalundefcoll, "HCALUndef");

	int HcalUndefSize = hcalundef.size();
	int NChEcalCore = checalcorecluster->getNumberOfElements();
	int NNeECALCoreEM = ecalnecore_EM.size();
	int NNeECALCoreNonEM =ecalnecore_NonEM.size();
	int NChHcalCore = chhcalcorecluster->getNumberOfElements();

	float DisHCALCluster[HcalUndefSize][NChHcalCore + NNeECALCoreEM + NNeECALCoreNonEM + NChEcalCore];
	float MinDisToHcalChCore = 1.0E10;
	float MinDisToEcalNeCore_EM = 1.0E10;
	float MinDisToEcalNeCore_NonEM = 1.0E10; 
	float MinDisToEcalChCore = 1.0E10;
	float HcalClusterDepth = 0; 

	for(int i4 = 0; i4 < HcalUndefSize; i4++)
	{
		Cluster * a_hcalUndef = hcalundef[i4];
		CluPos = a_hcalUndef->getPosition();
		HcalClusterDepth = DisSeedSurface(CluPos);

		MinDisToHcalChCore = 1.0E10;
		MinDisToEcalNeCore_EM = 1.0E10;
		MinDisToEcalNeCore_NonEM = 1.0E10;
		MinDisToEcalChCore = 1.0E10;

		for(int k4 = 0; k4 < NChHcalCore; k4++)
		{
			Cluster *a_hcalChcore = dynamic_cast<Cluster*>(chhcalcorecluster->getElementAt(k4));
			DisHCALCluster[i4][k4] = BushDis(a_hcalUndef, a_hcalChcore);
			if(DisHCALCluster[i4][k4] < MinDisToHcalChCore)
			{
				MinDisToHcalChCore = DisHCALCluster[i4][k4];
			}
		}

		for(int j4 = 0; j4 < NNeECALCoreEM; j4++)
		{
			Cluster * a_ecalneCore = ecalnecore_EM[j4];
			DisHCALCluster[i4][NChHcalCore + j4] = BushDis(a_hcalUndef, a_ecalneCore);  //Vec 
			if(DisHCALCluster[i4][NChHcalCore + j4] < MinDisToEcalNeCore_EM)
			{
				MinDisToEcalNeCore_EM = DisHCALCluster[i4][NChHcalCore + j4];
			}
		}

		for(int l4 = 0; l4 < NNeECALCoreNonEM; l4++)
		{
			Cluster * a_ecalneCore = ecalnecore_NonEM[l4]; 
			DisHCALCluster[i4][NChHcalCore + l4 + NNeECALCoreEM] = BushDis(a_hcalUndef, a_ecalneCore);  //Vec 
			if(DisHCALCluster[i4][NChHcalCore + l4 + NNeECALCoreEM] < MinDisToEcalNeCore_NonEM)
			{
				MinDisToEcalNeCore_NonEM = DisHCALCluster[i4][NChHcalCore + l4 + NNeECALCoreEM];
			}
		}

		for(int m4 = 0; m4 < NChEcalCore; m4++)
		{
			Cluster *a_ecalChcore = dynamic_cast<Cluster*>(checalcorecluster->getElementAt(m4));
			DisHCALCluster[i4][NChHcalCore + m4 + NNeECALCoreEM + NNeECALCoreNonEM] = BushDis(a_hcalUndef, a_ecalChcore);  //Vec 
			if(DisHCALCluster[i4][NChHcalCore + m4 + NNeECALCoreEM + NNeECALCoreNonEM] < MinDisToEcalChCore)
			{
				MinDisToEcalChCore = DisHCALCluster[i4][NChHcalCore + m4 + NNeECALCoreEM + NNeECALCoreNonEM];
			}
		}

		// cout<<"Depth && Min Hcal Distances: EcalChCore, HcalChCore, EcalNeCoreEM, EcalNeCore_NonEM, "<<HcalClusterDepth<<" : "<<MinDisToEcalChCore<<" : "<<MinDisToHcalChCore<<" : "<<MinDisToEcalNeCore_EM<<" : "<<MinDisToEcalNeCore_NonEM<<endl; 

		if(MinDisToHcalChCore < 30.0 + 0.02 * HcalClusterDepth)	//Really Tight Ones
		{
			hcalfrag_TBM_CH.push_back(a_hcalUndef);
		}
		// else if ( MinDisToHcalChCore > 200 && MinDisToEcalNeCore_EM > 200 && MinDisToEcalNeCore_NonEM > 200 && MinDisToEcalChCore > 200 && a_hcalUndef->getEnergy() > 2)
		else if ( (MinDisToHcalChCore > 100 + 0.1* HcalClusterDepth && a_hcalUndef->getEnergy() > 2) || (MinDisToHcalChCore > 50 + 0.05* HcalClusterDepth && a_hcalUndef->getEnergy() > 5)  )	
		{
			hcalnecore.push_back(a_hcalUndef);
		}
		else if ( HcalClusterDepth < 250 && ( MinDisToEcalNeCore_NonEM < 50 || MinDisToEcalNeCore_EM < 50 ) )	//Barrel & EndCap; shall we believe EM can be linked to Hcal clu?
		{
			if(MinDisToEcalNeCore_EM < 50)
			{
				hcalfrag_TBM_NE_EM.push_back(a_hcalUndef);
			}
			else 
			{
				hcalfrag_TBM_NE_NonEM.push_back(a_hcalUndef);
			}
		}
		else
		{
			hcalundef_iso.push_back(a_hcalUndef);
		}
	}	

	LCCollection *hcal_TBM_CH = ClusterVecColl(hcalfrag_TBM_CH);
	LCCollection *hcal_TBM_NE_EM = ClusterVecColl(hcalfrag_TBM_NE_EM);
	LCCollection *hcal_TBM_NE_NonEM = ClusterVecColl(hcalfrag_TBM_NE_NonEM);
	LCCollection *hcal_ne_core = ClusterVecColl(hcalnecore);
	LCCollection *hcal_iso = ClusterVecColl(hcalundef_iso);

	evtPP->addCollection(hcal_TBM_CH, "Hcal_TBM_CH");
	evtPP->addCollection(hcal_TBM_NE_EM, "Hcal_TBM_NE_EM");
	evtPP->addCollection(hcal_TBM_NE_NonEM, "Hcal_TBM_NE_NonEM");
	evtPP->addCollection(hcal_ne_core, "Hcal_NE_Core");
	evtPP->addCollection(hcal_iso, "Hcal_Iso");

	//Check distance with ecal cores: if attached to ecal core ends; 

	LCCollection *ecalpotentialbks = ClusterVecColl(ecalpotentialbackscattering);
	LCCollection *ecalnecorecluster_em = ClusterVecColl(ecalnecore_EM);
	LCCollection *ecalnecorecluster_nonem = ClusterVecColl(ecalnecore_NonEM);
	LCCollection *ecalfragcluster = ClusterVecColl(ecalfrag);
	LCCollection *ecalundefcluster = ClusterVecColl(ecalundef);
	LCCollection *ecalfragTBMCH=ClusterVecColl(ecalfrag_TBM_CH);
	LCCollection *ecalfragTBMNE_EM=ClusterVecColl(ecalfrag_TBM_NE_EM);
	LCCollection *ecalfragTBMNE_NonEM=ClusterVecColl(ecalfrag_TBM_NE_NonEM);
	LCCollection *ecalundefIso=ClusterVecColl(ecalundef_iso);

	evtPP->addCollection(ecalnecorecluster_em, "EcalCore_NE_EM");
	evtPP->addCollection(ecalnecorecluster_nonem, "EcalCore_NE_NonEM");
	evtPP->addCollection(ecalfragcluster, "EcalFrag");      
	evtPP->addCollection(ecalfragTBMCH, "Ecal_TBM_CH");
	evtPP->addCollection(ecalfragTBMNE_EM, "Ecal_TBM_NE_EM");
	evtPP->addCollection(ecalfragTBMNE_NonEM, "Ecal_TBM_NE_NonEM");
	evtPP->addCollection(ecalundefIso, "Ecal_Iso");
	evtPP->addCollection(ecalundefcluster, "EcalUndef");
	evtPP->addCollection(ecalpotentialbks, "Ecal_P_Bks");

	//Above is the full equivalent set of Input: EcalBushes & HcalBushes: now let's absorb those already tagged ones. 

	std::vector<Cluster* > chbush = ClusterAbsorbtion(chargedclustercore, hcalfrag_TBM_CH, 1000);
	chargedclustercore_abs = ClusterAbsorbtion(chbush, ecalfrag_TBM_CH, 1000);

	LCCollection *absorbedChCore = ClusterVecColl(chargedclustercore_abs);
	evtPP->addCollection(absorbedChCore, "ChargedCore");

	int N_ChartgedCore = chargedclustercore_abs.size();

	std::vector<Cluster* > ecal_em_core = ClusterAbsorbtion(ecalnecore_EM, ecalfrag_TBM_NE_EM, 1000);
	std::vector<Cluster* > em_core = ClusterAbsorbtion(ecal_em_core, hcalfrag_TBM_NE_EM, 1000);
	std::vector<Cluster* > ecal_nonem_core = ClusterAbsorbtion(ecalnecore_NonEM, ecalfrag_TBM_NE_NonEM, 1000);
	std::vector<Cluster* > nonem_core = ClusterAbsorbtion(ecal_nonem_core, hcalfrag_TBM_NE_NonEM, 1000);

	LCCollection *absorbedNeCore_EM = ClusterVecColl(em_core);
	LCCollection *absorbedNeCore_NonEM = ClusterVecColl(nonem_core);

	evtPP->addCollection(absorbedNeCore_EM, "EMCore");
	evtPP->addCollection(absorbedNeCore_NonEM, "NonEMCore");

	/*
	   std::cout<<"AAA Check"<<std::endl;

	   std::vector<std::string> TotalSubClass; 
	   TotalSubClass.push_back("EcalBushes");
	   TotalSubClass.push_back("HcalBushes");
	   TotalSubClass.push_back("ChargedCore");
	   TotalSubClass.push_back("EMCore");
	   TotalSubClass.push_back("NonEMCore");
	   TotalSubClass.push_back("Ecal_P_Bks");
	   TotalSubClass.push_back("Ecal_Iso");
	   TotalSubClass.push_back("EcalFrag");
	   TotalSubClass.push_back("Hcal_Iso");
	   TotalSubClass.push_back("Hcal_NE_Core");
	   CheckConsistence(evtPP, TotalSubClass);

	   std::cout<<"BBB Check"<<std::endl;
	 */

	//Re-create the Map between Tracks and Charged Core: which is stupid...

	// float DisMatrix_Track_CoreClu[NTrk][N_ChartgedCore];

	std::vector<Track*> MIPTracks; 
	std::vector<Track*> EMTracks; 
	std::vector<Track*> HadTracks; 
	std::vector<Track*> LowETracks; 	// No Cluster
	std::vector<Track*> UnBalanceTracks; 	// Cluster Energy >> Track Momentum
	std::vector<Track*> NoCluTracks;	// Energetic Track, but has no Cluster associated - might be in Cluster in eff range

	MIPTracks.clear();
	EMTracks.clear();
	HadTracks.clear();
	LowETracks.clear();
	UnBalanceTracks.clear();
	NoCluTracks.clear();

	LCCollection *arborrecoparticle_ch = new LCCollectionVec(LCIO::RECONSTRUCTEDPARTICLE);
	LCCollection *arborrecoparticle_ne = new LCCollectionVec(LCIO::RECONSTRUCTEDPARTICLE);
	LCCollection *arborchargedcorecluster = new LCCollectionVec(LCIO::CLUSTER);
	LCCollection *arborneutralcorecluster = new LCCollectionVec(LCIO::CLUSTER);
	arborchargedcorecluster->setFlag(Cluflag.getFlag());
	arborneutralcorecluster->setFlag(Cluflag.getFlag());

	//Non_Core_Neutral_Cluster	Push...
	//Core_Neutral_Cluster		Push...
	//
	//	Tight Absorb Coll
	//	Loose Absorb Coargedclustercore_absl

	for(int j5 = 0; j5 < NTrk; j5++)
	{
		Track* a_trk = SortedTracks[j5];

		int Track_Core_ID = 0;

		if(Map_Track_ClusterEnergy[a_trk] > 0)		// If already a ... trk	
		{
			currTrackType = Track_Type[a_trk];
			currTrkEn = Track_Energy[a_trk];
			float MinTrkCluDis = 1.0E10;
			int CloseCluIndex = -1;
			float CoreClusterEnergy = 0; 

			for(int k5 = 0; k5 < N_ChartgedCore; k5++)
			{
				Cluster * a_chargedcore_abs = chargedclustercore_abs[k5];
				float Dis = SimpleDisTrackClu(a_trk, a_chargedcore_abs);	
				// DisMatrix_Track_CoreClu[j5][k5] = Dis; 
				if(Dis < MinTrkCluDis && Dis > -0.1 && Map_Track_ClusterEnergy[a_trk] <= a_chargedcore_abs->getEnergy() )   //  ) && a_chargedcore_abs->getEnergy() < currTrkEn + 2*sqrt(currTrkEn) + 2)
				{
					MinTrkCluDis = Dis; 
					CloseCluIndex = k5; 
				}
			}

			// cout<<"ABS CluIndex, MinDis "<<CloseCluIndex<<" : "<<MinTrkCluDis<<" track energy "<<currTrkEn<<endl;

			if(CloseCluIndex > -1)
			{
				Cluster *a_close_tree = chargedclustercore_abs[CloseCluIndex];

				Track_Core_ID = ClusterFlag(a_close_tree, a_trk);

				if(BushTouchFlag.find(a_close_tree) == BushTouchFlag.end())
				{
					BushTouchFlag[a_close_tree] = currTrackType;
					FCMap_Track_ABSCluster[a_trk] = a_close_tree;
					CoreClusterEnergy = a_close_tree->getEnergy();
					if(CoreClusterEnergy > currTrkEn + 2*sqrt(currTrkEn) + 2)	//Treshold need to be adjustified
					{
						UnBalanceTracks.push_back(a_trk);
					}
					else if( Track_Core_ID == 110 )		//MatchedEM
					{
						EMTracks.push_back(a_trk);
						// Scan over tight region
					}
					else if(Track_Core_ID == 111)
					{
						EMTracks.push_back(a_trk);
						// Scan over both tight/loose region
						// tight: Non EM && Smaller distances...
					}
					else if( Track_Core_ID == 130 )		//Matchched
					{
						MIPTracks.push_back(a_trk);
					}
					else if( Track_Core_ID == 131 )
					{
						MIPTracks.push_back(a_trk);
						// Scan over back region
					}
					else if( Track_Core_ID == 211)		//Matched
					{
						HadTracks.push_back(a_trk);
						// Scan over tight region, according to Track_Clu distance & MCCore_Clu distance
					}
					else if( Track_Core_ID == 212)	// 211, all non
					{
						HadTracks.push_back(a_trk);
						// Scan over loose region
					}
					else 
					{
						cout<<"Fragment tagged "<<Track_Core_ID<<endl; 
						//Frag & UnBalance, need to re-tag the Core Cluster.
					}
				}
				else if(Track_Energy[a_trk] > 2)
				{
					NoCluTracks.push_back(a_trk);
				}
				else
				{
					LowETracks.push_back(a_trk);
				}
				// cout<<j5<<"th "<<Track_Core_ID<<" Track with Energy "<<currTrkEn<<" has a Core Cluster with Energy "<<Map_Track_ClusterEnergy[a_trk]<<" : "<<CoreClusterEnergy<<endl;
			}
		}
		else if(Track_Energy[a_trk] > 2)
		{
			NoCluTracks.push_back(a_trk);
		}
		else
		{
			LowETracks.push_back(a_trk);
		}

		//Shall I veto some crazy track...

		ReconstructedParticleImpl * chargeparticle = new ReconstructedParticleImpl();
		chargeparticle->setEnergy( Track_Energy[a_trk] );
		chargeparticle->setCharge(a_trk -> getOmega()/fabs(a_trk -> getOmega()));
		HelixClass * currHelix = new HelixClass();
		currHelix->Initialize_Canonical(a_trk->getPhi(), a_trk -> getD0(), a_trk -> getZ0(), a_trk -> getOmega(), a_trk->getTanLambda(), BField);
		chargeparticle->setMomentum( currHelix->getMomentum() );
		chargeparticle->addTrack( a_trk );
		if( FCMap_Track_ABSCluster[a_trk] )
		{
			ClusterImpl * chargedarborcluster =  NaiveCluImpl(FCMap_Track_ABSCluster[a_trk]);

			chargeparticle->addCluster(chargedarborcluster);

			arborchargedcorecluster->addElement(chargedarborcluster);
		}
		arborrecoparticle_ch->addElement(chargeparticle);
		ChCoreID[chargeparticle] = Track_Core_ID;
	}

	evtPP->addCollection( arborchargedcorecluster, "ClusterChargedCore" );
	evtPP->addCollection( arborrecoparticle_ch, "ArborChargedCore" );

	// cout<<"Track_Core Pair Classification: EM, MIP, Had, Unbalance, NoCluTracks, LowETracks "<<EMTracks.size()<<", "<<MIPTracks.size()<<", "<<HadTracks.size()<<", "<<UnBalanceTracks.size()<<", "<<NoCluTracks.size()<<", "<<LowETracks.size()<<" =? "<<NTrk<<" =?? "<<arborrecoparticle_ch->getNumberOfElements()<<endl<<endl; 


	//~~~~~~~~~~~~~~~~~~~~~

	//Reabsorbtion also defined over here...

	//Cores:
	int NEMCore = em_core.size();
	int NNonEMCore = nonem_core.size();
	int NNonEMHcalCore = hcalnecore.size();
	int NNeutralCore = NEMCore + NNonEMCore + NNonEMHcalCore;
	float NAMom[3] = {0, 0, 0};
	TVector3 BushSeedPos; 

	// std::map<Cluster, int> ToucedCluster_NC; 
	int tmp_ClusterTypeID = 0; 

	int NBKS = ecalpotentialbackscattering.size();
	int NEcalIso = ecalundef_iso.size();
	int NHcalIso = hcalundef_iso.size();
	int NEcalFrag = ecalfrag.size();
	int NFrags = NEcalIso + NHcalIso + NEcalFrag + NBKS;

	//Frags:

	for(int j7 = 0; j7 < NFrags; j7++)
	{
		Cluster * a_clu(0);

		if(j7 < NHcalIso)	//from back to front...
		{
			a_clu = hcalundef_iso[j7];
			tmp_ClusterTypeID = 10;
		}
		else if(j7 < NEcalFrag + NHcalIso)
		{
			a_clu = ecalfrag[j7 - NHcalIso];
			tmp_ClusterTypeID = 11;
		}
		else if(j7 < NEcalFrag + NHcalIso + NEcalIso)
		{
			a_clu = ecalundef_iso[j7 - NEcalFrag - NHcalIso];
			tmp_ClusterTypeID = 12;
		}
		else
		{
			a_clu = ecalpotentialbackscattering[j7 - NEcalFrag - NHcalIso - NEcalIso];
			tmp_ClusterTypeID = 14;
		}
		non_chargedclustercore.push_back(a_clu);
		ClusterType_1stID[a_clu] = tmp_ClusterTypeID;
	}


	for(int j6 = 0; j6< NNeutralCore; j6++)	
	{
		Cluster * a_clu(0); 

		if(j6 < NEMCore)
		{
			a_clu = em_core[j6];
			tmp_ClusterTypeID = 1;
		}
		else if(j6 < NEMCore + NNonEMCore)
		{
			a_clu = nonem_core[j6 - NEMCore];
			tmp_ClusterTypeID = 2;
		}
		else
		{
			a_clu = hcalnecore[j6 - NEMCore - NNonEMCore];
			tmp_ClusterTypeID = 3; 
		}

		non_chargedclustercore.push_back(a_clu);
		ClusterType_1stID[a_clu] = tmp_ClusterTypeID; 

		BushSeedPos = a_clu->getPosition();

		ReconstructedParticleImpl * neutralparticle = new ReconstructedParticleImpl();
		neutralparticle->setEnergy( a_clu->getEnergy() );
		neutralparticle->setMass( 0.0 );
		neutralparticle->setCharge( 0.0 );
		NAMom[0] = CluEnergy*1.0/BushSeedPos.Mag()*BushSeedPos.X();
		NAMom[1] = CluEnergy*1.0/BushSeedPos.Mag()*BushSeedPos.Y();
		NAMom[2] = CluEnergy*1.0/BushSeedPos.Mag()*BushSeedPos.Z();
		neutralparticle->setMomentum( NAMom );
		ClusterImpl * neutralarborcluster =  NaiveCluImpl(a_clu);
		neutralparticle->addCluster(neutralarborcluster);
		arborneutralcorecluster->addElement(neutralarborcluster);
		arborrecoparticle_ne->addElement(neutralparticle);
	}

	// LCCollection *nn_core = ClusterVecColl(NNCore);

	// evtPP->addCollection(nn_core, "ClusterNeutralCore");

	evtPP->addCollection( arborneutralcorecluster, "ClusterNeutralCore");
	evtPP->addCollection( arborrecoparticle_ne, "ArborNeutralCore");


	//Calculate the distance between helix // and cluster...






	//loop over ChargedCore, For each Core, create an array of near by clusters, sorted according to effecive distance.

	/*

	// std::vector<Cluster* > nebush = ClusterAbsorbtion(, hcalfrag_TBM_CH, 1000);

	std::vector<std::string> ChargedSubClass; 
	ChargedSubClass.push_back("ChargedCore");
	ChargedSubClass.push_back("ECALCORE_CH");
	ChargedSubClass.push_back("HCALCORE_CH");
	ChargedSubClass.push_back("Ecal_TBM_CH");
	ChargedSubClass.push_back("Hcal_TBM_CH");
	CheckConsistence(evtPP, ChargedSubClass);

	//std::map<Track*, std::vector<int>> something: mapping between track to its clusters (supercluster...)

	//Verification

	std::vector<std::string> EcalSubClass;
	std::vector<std::string> HcalSubClass;	

	EcalSubClass.push_back("EcalBushes");
	EcalSubClass.push_back("ECALCORE_CH"); 
	EcalSubClass.push_back("EcalCore_NE_EM");
	EcalSubClass.push_back("EcalCore_NE_NonEM");
	EcalSubClass.push_back("EcalFrag");
	EcalSubClass.push_back("Ecal_TBM_CH");
	EcalSubClass.push_back("Ecal_TBM_NE_EM");
	EcalSubClass.push_back("Ecal_TBM_NE_NonEM");
	EcalSubClass.push_back("Ecal_Iso");	
	EcalSubClass.push_back("Ecal_P_Bks");
	CheckConsistence(evtPP, EcalSubClass);

	HcalSubClass.push_back("HcalBushes");
	HcalSubClass.push_back("HCALCORE_CH");
	HcalSubClass.push_back("Hcal_NE_Core");
	HcalSubClass.push_back("Hcal_TBM_CH");
	HcalSubClass.push_back("Hcal_TBM_NE_EM");
	HcalSubClass.push_back("Hcal_TBM_NE_NonEM");
	HcalSubClass.push_back("Hcal_Iso");
	CheckConsistence(evtPP, HcalSubClass);

	 */

}

void BushConnect::ParticleReco( LCEvent * evtPP )
{

	//	LCCollection *arborrecoparticle_ch_merged = new LCCollectionVec(LCIO::RECONSTRUCTEDPARTICLE);
	//	LCCollection *arborrecoparticle_ne_merged = new LCCollectionVec(LCIO::RECONSTRUCTEDPARTICLE);
	LCCollection *arborrecoparticle = new LCCollectionVec(LCIO::RECONSTRUCTEDPARTICLE);
	LCCollection *mergedclu_ch = new LCCollectionVec(LCIO::CLUSTER);
	LCCollection *mergedclu_ne = new LCCollectionVec(LCIO::CLUSTER);
	mergedclu_ch->setFlag(Cluflag.getFlag());
	mergedclu_ne->setFlag(Cluflag.getFlag());

	LCCollection * ChargedCore = evtPP->getCollection("ArborChargedCore");

	int NChargedObj = ChargedCore->getNumberOfElements();

	int NNeutralCluster = non_chargedclustercore.size();

	double DisMatrix_Core_Neutral[NChargedObj][NNeutralCluster][2];		//Define different types of distances; 

	float TotalChEn = 0;
	float TotalNeEn = 0; 

	Track * a_chargedTrk(0); 
	Cluster * a_chargedClu(0), *a_NeCandiClu(0); 
	float CluDepth = 0;
	std::map<Cluster*, double> CluDepthMap; 
	CluDepthMap.clear();
	int currChargeCoreType = 0;  
	TVector3 CluPos; 

	// Per Track usage...
	std::vector<Cluster*> loosecandicluster; 
	std::vector<Cluster*> tightcandicluster;		//Muon potential candi?
	std::vector<Cluster*> mergedcluster; 			//tmp for each charged P
	std::vector<Cluster*> chargedclustercore_merged; 	//overall

	chargedclustercore_merged.clear();

	std::vector<double> reftightdis; 
	std::vector<double> refloosedis; 

	std::map<Cluster*, int> NNCTouchFlag; 
	std::vector<Track*> SecondIterTracks;
	SecondIterTracks.clear();

	for(int i = 0; i < NChargedObj; i++)
	{
		ReconstructedParticle * a_recoP_ch = dynamic_cast<ReconstructedParticle*>(ChargedCore->getElementAt(i));

		loosecandicluster.clear();
		tightcandicluster.clear();
		mergedcluster.clear();
		reftightdis.clear();
		refloosedis.clear();

		a_chargedTrk = a_recoP_ch->getTracks()[0];

		currChargeCoreType = ChCoreID[a_recoP_ch];

		float CurrClusterEnergy = 0;

		float CurrTrackEnergy = Track_Energy[a_chargedTrk];

		if(a_recoP_ch->getClusters().size() != 0)
		{
			a_chargedClu = a_recoP_ch->getClusters()[0];
			CurrClusterEnergy = a_chargedClu->getEnergy();
			//	if( CurrClusterEnergy < CurrTrackEnergy + 2*sqrt(CurrTrackEnergy) + 4 )
			mergedcluster.push_back(a_chargedClu);		//Actually can use this chance to question if previous energy are balance...
		}

		float MinDisToNoClusterTrk = 1.0E10; 

		for(int j = 0; j < NNeutralCluster; j++)
		{
			a_NeCandiClu = non_chargedclustercore[j];

			CluPos = a_NeCandiClu->getPosition();
			CluDepth = DisSeedSurface(CluPos);
			CluDepthMap[a_NeCandiClu] = CluDepth; 	

			float tmpPiWeight = 0;
			if(a_NeCandiClu->getEnergy() > 3)
				tmpPiWeight = 0.1;
			else if(a_NeCandiClu->getEnergy() > 1.5)
				tmpPiWeight = 0.05; 

			if( ClusterType_1stID[a_NeCandiClu] == 1 )   continue; 
			// && currChargeCoreType != 111 ) continue;	//already the photon protection

			for(int k = 0; k < 2; k++)
			{
				DisMatrix_Core_Neutral[i][j][k] = 1.0E9;
			}

			if(CurrClusterEnergy > 1E-6)	//put by hand...
			{
				DisMatrix_Core_Neutral[i][j][0] = BushDis(a_chargedClu, a_NeCandiClu);
			}
			DisMatrix_Core_Neutral[i][j][1] = SimpleDisTrackClu(a_chargedTrk, a_NeCandiClu);



			if( NNCTouchFlag.find(a_NeCandiClu) == NNCTouchFlag.end() && ( currChargeCoreType == 0 || DisMatrix_Core_Neutral[i][j][0] < 1000 ))
			{			
				if( currChargeCoreType == 130 )			//Matched Muon, should ignore
				{
					if( DisMatrix_Core_Neutral[i][j][1] < 0.2*CluDepth && CluDepth > 200  )	//&& FD?
					{
						tightcandicluster.push_back(a_NeCandiClu);
						reftightdis.push_back( DisMatrix_Core_Neutral[i][j][1] );	//dependence on Cluster Flag & Clu Depth. use some more fancy sort algorithm...
					}
				}
				else if( currChargeCoreType == 131 )
				{
					if( DisMatrix_Core_Neutral[i][j][1] < 0.3*CluDepth && CluDepth > 150 )
					{
						tightcandicluster.push_back(a_NeCandiClu);
						reftightdis.push_back( DisMatrix_Core_Neutral[i][j][1] );	
					}
					else if( DisMatrix_Core_Neutral[i][j][1] < 0.5*CluDepth && CluDepth > 100 )
					{
						loosecandicluster.push_back(a_NeCandiClu);
						refloosedis.push_back( DisMatrix_Core_Neutral[i][j][1] );
					}
				}	
				else if( currChargeCoreType == 110  )		// Electron
				{
					if( DisMatrix_Core_Neutral[i][j][0] < 0.15*CluDepth + 15 )
					{
						tightcandicluster.push_back(a_NeCandiClu);
						reftightdis.push_back( DisMatrix_Core_Neutral[i][j][0] );
					}
				}			
				else if( currChargeCoreType == 111 )		// look behind... might be pion...
				{
					if( DisMatrix_Core_Neutral[i][j][0] < 0.1*CluDepth + 15 && DisMatrix_Core_Neutral[i][j][1] < 0.1*CluDepth + 10 )	//Define Brems Photon region for correct
					{
						tightcandicluster.push_back(a_NeCandiClu);
						if(DisMatrix_Core_Neutral[i][j][0] < DisMatrix_Core_Neutral[i][j][1])	// not fully adequate.
						{
							reftightdis.push_back( DisMatrix_Core_Neutral[i][j][0] );
						}
						else
						{
							reftightdis.push_back( DisMatrix_Core_Neutral[i][j][1] );
						}
					}
					else if( DisMatrix_Core_Neutral[i][j][0] < 0.2*CluDepth + 15 || DisMatrix_Core_Neutral[i][j][1] < 0.2*CluDepth + 15  )
					{	
						loosecandicluster.push_back(a_NeCandiClu);

						if(DisMatrix_Core_Neutral[i][j][0] < DisMatrix_Core_Neutral[i][j][1])   // not fully adequate.
						{
							refloosedis.push_back( DisMatrix_Core_Neutral[i][j][0] );
						}
						else
						{
							refloosedis.push_back( DisMatrix_Core_Neutral[i][j][1] );
						}
					}
				}
				else if( currChargeCoreType == 211 )	//Main Cluster distance oriented
				{
					if(DisMatrix_Core_Neutral[i][j][0] < 0.2*CluDepth)
					{
						tightcandicluster.push_back(a_NeCandiClu);
						reftightdis.push_back( DisMatrix_Core_Neutral[i][j][0] );
					}
				}
				else if( currChargeCoreType == 212 )	//Non_Matched
				{
					if(DisMatrix_Core_Neutral[i][j][0] < 10 + (0.3 - tmpPiWeight)*CluDepth )	//Energy Dependence...
					{
						tightcandicluster.push_back(a_NeCandiClu);
						reftightdis.push_back( DisMatrix_Core_Neutral[i][j][0] );
					}
					else if(DisMatrix_Core_Neutral[i][j][0] < 10 + (0.4 - 2*tmpPiWeight)*CluDepth || DisMatrix_Core_Neutral[i][j][1] < 20 + (0.5 - 3*tmpPiWeight)*CluDepth )
					{
						loosecandicluster.push_back(a_NeCandiClu);
						refloosedis.push_back( DisMatrix_Core_Neutral[i][j][0] );
					}
				}
				else if( currChargeCoreType == 0 )	//
				{
					if(CluDepth < 20)
					{
						if(DisMatrix_Core_Neutral[i][j][1] < MinDisToNoClusterTrk)	//Tag minimal distance cluster... and see if it can be potentially linked.
						{
							MinDisToNoClusterTrk = DisMatrix_Core_Neutral[i][j][1];
						}
						if( MinDisToNoClusterTrk < 300 )	//some hard cut
						{
							tightcandicluster.push_back(a_NeCandiClu);
							reftightdis.push_back( DisMatrix_Core_Neutral[i][j][1] );
						}
					}
				}
				else
				{
					cout<<"Over balanced/Un matched/defined case: "<<currChargeCoreType<<endl; 
				}
			}
		}

		if( currChargeCoreType == 130 )
		{
			for(unsigned int i1 = 0; i1 < tightcandicluster.size(); i1++)
			{
				Cluster* a_clu = tightcandicluster[i1];
				if( ClusterType_1stID[a_clu] >= 10 ) //  && CurrClusterEnergy < CurrTrackEnergy + 1.0*sqrt(CurrTrackEnergy) )	//Frags...
				{
					mergedcluster.push_back( a_clu );		
					CurrClusterEnergy += a_clu->getEnergy();
				}
				else if( ClusterType_1stID[a_clu] < 10 && CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 2.0*sqrt(CurrTrackEnergy) )
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}
		}
		else if( currChargeCoreType == 131 )
		{
			for(unsigned int i1 = 0; i1 < tightcandicluster.size(); i1++)
			{
				Cluster* a_clu = tightcandicluster[i1];
				if( ClusterType_1stID[a_clu] >= 10 || (ClusterType_1stID[a_clu] < 10 && CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 1.0*sqrt(CurrTrackEnergy))  ) 
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}

			//Maybe Some ID over here?...	//layers & numbers...	//BS Case ID

			for(unsigned int i2 = 0; i2 < loosecandicluster.size(); i2++)
			{
				Cluster* a_clu = loosecandicluster[i2];
				if( ClusterType_1stID[a_clu] >= 10 || CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 2.0*sqrt(CurrTrackEnergy) )       //Frags...Or some minmal hit cut
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}
		}
		else if( currChargeCoreType == 110 )
		{
			for(unsigned int i1 = 0; i1 < tightcandicluster.size(); i1++)
			{
				Cluster* a_clu = tightcandicluster[i1];
				if( ClusterType_1stID[a_clu] >= 10 || (ClusterType_1stID[a_clu] < 10 && CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 0.5*sqrt(CurrTrackEnergy))  )
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}
		}
		else if( currChargeCoreType == 111 )
		{
			for(unsigned int i1 = 0; i1 < tightcandicluster.size(); i1++)
			{
				Cluster* a_clu = tightcandicluster[i1];
				if( ClusterType_1stID[a_clu] >= 10 || (ClusterType_1stID[a_clu] < 10 && CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 0.5*sqrt(CurrTrackEnergy))  )
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}

			for(unsigned int i2 = 0; i2 < loosecandicluster.size(); i2++)
			{
				Cluster* a_clu = loosecandicluster[i2];
				if( ClusterType_1stID[a_clu] >= 10 || fabs(CurrClusterEnergy + a_clu->getEnergy() - CurrTrackEnergy) < fabs(CurrClusterEnergy - CurrTrackEnergy) )       //Frags...Or some minmal hit cut
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}	
		}
		else if( currChargeCoreType == 211 )	// Matched
		{
			for(unsigned int i1 = 0; i1 < tightcandicluster.size(); i1++)
			{
				Cluster* a_clu = tightcandicluster[i1];
				if( ClusterType_1stID[a_clu] >= 10 || (ClusterType_1stID[a_clu] < 10 && CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 2.0*sqrt(CurrTrackEnergy))  )
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}
		}
		else if( currChargeCoreType == 212)
		{
			for(unsigned int i1 = 0; i1 < tightcandicluster.size(); i1++)
			{
				Cluster* a_clu = tightcandicluster[i1];
				if( ClusterType_1stID[a_clu] >= 10 || (ClusterType_1stID[a_clu] < 10 && CurrClusterEnergy + a_clu->getEnergy() < CurrTrackEnergy + 2.0*sqrt(CurrTrackEnergy))  )
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}

			for(unsigned int i2 = 0; i2 < loosecandicluster.size(); i2++)
			{
				Cluster* a_clu = loosecandicluster[i2];
				if( ClusterType_1stID[a_clu] >= 10 || fabs(CurrClusterEnergy + a_clu->getEnergy() - CurrTrackEnergy) < fabs(CurrClusterEnergy - CurrTrackEnergy) )       //Frags...Or some minmal hit cut
				{
					mergedcluster.push_back( a_clu );
					CurrClusterEnergy += a_clu->getEnergy();
				}
			}
		}
		else if( currChargeCoreType == 0 && reftightdis.size() > 0)
		{
			float mindis = 1.0E10;
			int minindex = 0; 

			for(unsigned int i1 = 0; i1 < reftightdis.size(); i1 ++)
			{
				if(reftightdis[i1] < mindis)
				{
					mindis = reftightdis[i1];
					minindex = i1; 
				}
			}

			Cluster* a_clu = tightcandicluster[minindex];	// Only 1? ...

			mergedcluster.push_back( a_clu );
		}
		else
		{
			cout<<"No_match"<<endl; 
		}

		// cout<<"MCSize  "<<mergedcluster.size()<<endl; 
		float CHCluEnergy = 0;
		// float PFNEMom[3] = {0, 0, 0};

		// Should add some energy selection critieria

		for(int is = 0; is < mergedcluster.size(); is++)
		{       
			Cluster* a_TBM_clu = mergedcluster[is]; 
			CHCluEnergy += a_TBM_clu->getEnergy();
		}

		//if( fabs(CHCluEnergy - CurrTrackEnergy) < 2 + 2.5*sqrt(CurrTrackEnergy) || currChargeCoreType == 130 || currChargeCoreType == 131   ) 	// Else push to recycling ...

		// Exist the case where two closed track cluster got merged. so the energy of one of them will be limited. 

		if( !( CHCluEnergy < 1.0 && CurrTrackEnergy > 12 ) || Track_Type[a_chargedTrk]%10 == 1 )	// HQ Tracks, always believe...
		{

			for(int i2 = 0; i2 < mergedcluster.size(); i2++)
			{
				Cluster* a_TBM_clu = mergedcluster[i2];
				NNCTouchFlag[a_TBM_clu]	= 2; 		// can make use of this intereting flag...
				//CHCluEnergy += a_TBM_clu->getEnergy();
			}

			int charge = a_chargedTrk -> getOmega()/fabs(a_chargedTrk -> getOmega());
			ReconstructedParticleImpl * chargeparticle = new ReconstructedParticleImpl();
			float RecoEn = CurrTrackEnergy;
                        //Clustermerging
                        ClusterImpl * chclustermerged =  NaiveMergeClu(mergedcluster);
                        mergedclu_ch->addElement(chclustermerged);
                        chargeparticle->addCluster(chclustermerged);
                        chargedclustercore_merged.push_back(chclustermerged);

                        int currChargeCoreType2;
                        currChargeCoreType2 = ClusterFlag(chclustermerged, a_chargedTrk); 			

			if (!(currChargeCoreType2 == 130 || currChargeCoreType2 == 131) && CurrTrackEnergy > 100 && CurrTrackEnergy > 10 * CHCluEnergy) RecoEn = CHCluEnergy;
			chargeparticle->setEnergy( RecoEn );
			chargeparticle->setCharge(charge);
			HelixClass * currHelix = new HelixClass();
			currHelix->Initialize_Canonical(a_chargedTrk->getPhi(), a_chargedTrk -> getD0(), a_chargedTrk -> getZ0(), a_chargedTrk -> getOmega(), a_chargedTrk->getTanLambda(), BField);
			chargeparticle->setMomentum( currHelix->getMomentum() );
			chargeparticle->addTrack( a_chargedTrk );

			TotalChEn += CurrTrackEnergy;

			int flagEnergyFlow = 0;

			if(currChargeCoreType2 == 130 || currChargeCoreType2 == 131)
			{
				chargeparticle->setType( int(-13*charge) );
			}
			else if(currChargeCoreType2 == 110 || currChargeCoreType2 == 111)
			{
				chargeparticle->setType( int(-11*charge) );
				if(CHCluEnergy > CurrTrackEnergy + 0.5*sqrt(CurrTrackEnergy) + 1)
				{
					flagEnergyFlow = 1; 
				}
			}
			else
			{
				chargeparticle->setType( int(211*charge) );
				if(CHCluEnergy > CurrTrackEnergy + 1.5*sqrt(CurrTrackEnergy) + 1)
				{
					flagEnergyFlow = 2;
				}
			}

			//Energy Flow Procedure

			if( flagEnergyFlow )
			{
				ReconstructedParticleImpl * a_Ef_Ne_particle = new ReconstructedParticleImpl();
				a_Ef_Ne_particle->setEnergy( CHCluEnergy - CurrTrackEnergy );
				TVector3 corePos = chclustermerged->getPosition();
				float WFactor = (CHCluEnergy - CurrTrackEnergy)/corePos.Mag(); 
				float PFNEMom[3] = {WFactor*corePos.X(), WFactor*corePos.Y(), WFactor*corePos.Z()};
				a_Ef_Ne_particle->setMomentum(PFNEMom);
				a_Ef_Ne_particle->setMass( 0.0 );
				a_Ef_Ne_particle->setCharge( 0.0 );
				a_Ef_Ne_particle->setType(310);		//Define as Klong
				arborrecoparticle->addElement(a_Ef_Ne_particle);

				a_Ef_Ne_particle->addCluster(chclustermerged);	//Maybe lead to crush
				cout<<"Energy Flow Neutral Tagged "<<CHCluEnergy - CurrTrackEnergy<<endl; 
			}
			arborrecoparticle->addElement(chargeparticle);

		}
		else	// push non valid tracks, etc to second iteration, as those for PreInteracting ones
		{
			SecondIterTracks.push_back(a_chargedTrk);
		}	

	}

	evtPP->addCollection(mergedclu_ch, "ArborCharged");

	// Define some absorbtion?? Hit level, or...? 

	LCCollection * IterFrag_ECAL = evtPP->getCollection("Ecal_IterFrag");
	LCCollection * IterFrag_HCAL = evtPP->getCollection("Hcal_IterFrag");

	std::vector<Cluster*> Ab_or_veto_clu;
	std::vector<Cluster*> BBCore; 
	Ab_or_veto_clu.clear();
	BBCore.clear();

	for(int k0 = 0; k0 < IterFrag_ECAL->getNumberOfElements(); k0++)
	{
		Cluster * a_clu = dynamic_cast<Cluster*>(IterFrag_ECAL->getElementAt(k0));
		Ab_or_veto_clu.push_back(a_clu);
	}

	for(int k1 = 0; k1 < IterFrag_HCAL->getNumberOfElements(); k1++)
	{
		Cluster * b_clu = dynamic_cast<Cluster*>(IterFrag_HCAL->getElementAt(k1));
		Ab_or_veto_clu.push_back(b_clu);
	}

	std::map<Cluster*, float> CoreEn; 
	CoreEn.clear();

	/*
	   std::map<Cluster*, float> ChargeCoreDis; 	
	   ChargeCoreDis.clear();

	   std::map<Cluster*, float> NeutralCoreDis;
	   NeutralCoreDis.clear();
	 */

	std::vector<Cluster*> UnclasifiedCC; 
	UnclasifiedCC.clear();

	// Merging cores which are back to back...




	for(int k6 = 0; k6 < NNeutralCluster; k6 ++)
	{
		Cluster * c_clu = non_chargedclustercore[k6];
		float tmpDis = 1.0E10; 
		float MinDisToChCore = 1.0E10;

		if( NNCTouchFlag.find(c_clu) == NNCTouchFlag.end() )
		{
			if( ClusterType_1stID[c_clu] < 10 )
			{
				BBCore.push_back(c_clu);
				CoreEn[c_clu] = c_clu->getEnergy();
			}
			else
			{
				MinDisToChCore = 1.0E10; 
				for(int s = 0; s < mergedclu_ch->getNumberOfElements(); s ++ )
				{
					Cluster* a_chcore = dynamic_cast<Cluster*>(mergedclu_ch->getElementAt(s));
					tmpDis = BushDis(a_chcore, c_clu);
					if(tmpDis < MinDisToChCore)
					{
						MinDisToChCore = tmpDis;
					}
				}

				if( MinDisToChCore > 10 + 0.04*CluDepthMap[c_clu] && c_clu->getEnergy() > 0.05 + 0.001*CluDepthMap[c_clu]) // && c_clu->getCalorimeterHits().size() > 3)
				{
					BBCore.push_back(c_clu);
					CoreEn[c_clu] = c_clu->getEnergy();
				}
				else // if( !(MinDisToChCore < 10 + 0.02*CluDepthMap[c_clu] && c_clu->getEnergy() < 0.2 + 0.001*CluDepthMap[c_clu]) )
				{
					UnclasifiedCC.push_back(c_clu);
				}
			}
		}
	}

	// InterLink BBCore according to 

	// ~~~~~~~~~~~~~

	for(int k7 = 0; k7 < int(UnclasifiedCC.size()); k7++)
	{
		Cluster * d_clu = UnclasifiedCC[k7];
		float tmpDis = 1.0E10;
		float MinDisToBBCore = 1.0E10;
		for(int k8 = 0; k8 < int(BBCore.size()); k8++)
		{
			Cluster* a_bbcore = BBCore[k8];
			tmpDis = BushDis(a_bbcore, d_clu);
			if(tmpDis < MinDisToBBCore)
			{
				MinDisToBBCore = tmpDis; 
			}
		}
		if(MinDisToBBCore > 50 && d_clu->getEnergy() > 0.2 + 0.002*CluDepthMap[d_clu]) // && d_clu->getCalorimeterHits().size() > 3)
		{
			BBCore.push_back(d_clu);
		}
		else
		{
			Ab_or_veto_clu.push_back(d_clu);	
		}
	}

	int NAb_or_veto = Ab_or_veto_clu.size();

	std::vector<Cluster* > neutralfragTBA; 
	neutralfragTBA.clear();

	for(int z = 0; z < NAb_or_veto; z++)
	{
		float DisToChCore = 1.0E10; 
		float DisToNeCore = 1.0E10; 
		float tmpDis = 1.0E10; 
		int CloseNeCoreIndex = -10; 
		Cluster * a_cand = Ab_or_veto_clu[z];
		TVector3 currPos = a_cand->getPosition();
		float DepthFrag = DisSeedSurface(currPos);

		for(int s = 0; s < mergedclu_ch->getNumberOfElements(); s ++ )
		{
			Cluster* a_chcore = dynamic_cast<Cluster*>(mergedclu_ch->getElementAt(s));
			tmpDis = BushDis(a_chcore, a_cand);			
			if(tmpDis < DisToChCore)
			{
				DisToChCore = tmpDis; 
			}
		}
		for(int t = 0; t < BBCore.size(); t ++)
		{
			Cluster * a_necore = BBCore[t];
			tmpDis = BushDis(a_necore, a_cand);
			if(tmpDis < DisToNeCore)
			{
				DisToNeCore = tmpDis;
				CloseNeCoreIndex = t; 
			}
		}

		if( DisToNeCore < 1.0 * DisToChCore && DisToNeCore < 100 + 0.1*DepthFrag )	//&& DisToNeCore < 110 + Depth * sth?
		{
			neutralfragTBA.push_back(a_cand);
			// CoreEn[BBCore[CloseNeCoreIndex]] += a_cand->getEnergy();
		}
	}

	//Should add a serious merge...
	// Shall we believe the charged part already provide enough information for ID...


	std::vector<Cluster*> ECALCluCore;
	std::vector<Cluster*> HCALCluCore;
	ECALCluCore.clear();
	HCALCluCore.clear();
	std::map<Cluster*, TVector3> CoG_Clu; 
	CoG_Clu.clear();
	TVector3 CluCoGE, CluCoGH; 

	for(int i8 = 0; i8 < BBCore.size(); i8++)
	{
		Cluster * a_clu = BBCore[i8];
		if(CluDepthMap[a_clu] < 100)    //ECAL Clu
			ECALCluCore.push_back(a_clu);
		else
			HCALCluCore.push_back(a_clu);
		CoG_Clu[a_clu] = ClusterCoG(a_clu);
	}

	for(int i9 = 0; i9 < ECALCluCore.size(); i9++)		//Should sort according to energy...
	{
		Cluster * a_clu_E = ECALCluCore[i9];
		CluCoGE = CoG_Clu[a_clu_E];
		for(int j9 = 0; j9 < HCALCluCore.size(); j9++)
		{
			Cluster * a_clu_H = HCALCluCore[j9];
			if( BushDis( a_clu_E, a_clu_H ) < 200 )
			{
				CluCoGH = CoG_Clu[a_clu_H];	
				TVector3 Diff = CluCoGH - CluCoGE; 
				if( Diff.Angle(CluCoGE) < 0.15 || Diff.Angle(CluCoGH) < 0.15 )
				{
					cout<<"ECAL Core and HCAL Core, Should be connect"<< BushDis(a_clu_E, a_clu_H) <<" : "<< Diff.Angle(CluCoGE) <<endl; 
				}
			}
		}
	}

        std::vector<Cluster* > finalneutralbush = ClusterAbsorbtion(BBCore, neutralfragTBA, 1000);
        //cout<<"verify size "<<BBCore.size()<<" ?= "<<finalneutralbush.size()<<endl; 

        std::map<Cluster*, float> FNCluEn;
        FNCluEn.clear();
        for(int p = 0; p < int(finalneutralbush.size()); p++)
        {
                FNCluEn[ finalneutralbush[p]  ] = finalneutralbush[p]->getEnergy();
        }

        if(1 == 1)
        {
                LCCollection * h_IsoHits = evtPP->getCollection("HcalIsoHits");
                int NHit = h_IsoHits->getNumberOfElements();
                float MinHitDisToCluster_Neutral = 1.0E10;
                float MinHitDisToCluster_Charge = 1.0E10;
                float tmpHitCluDis = 0;
                TVector3 a_hitPos;
                TVector3 a_cluPos;
                int CluIndex = -1;

                for(int s = 0; s < NHit; s++)
                {
                        CalorimeterHit * a_hit = dynamic_cast<CalorimeterHit*>( h_IsoHits->getElementAt(s) );
                        MinHitDisToCluster_Neutral = 1.0E10;
                        MinHitDisToCluster_Charge = 1.0E10;
                        a_hitPos = a_hit->getPosition();
                        CluIndex = -1;
                        for(int p = 0; p < int(finalneutralbush.size()); p++)
                        {
                                Cluster * a_clu = finalneutralbush[p];
                                a_cluPos = a_clu->getPosition();
                                if( (a_cluPos - a_hitPos).Mag() < 600 )
                                {
                                        tmpHitCluDis = DisPointToBush( a_hitPos, a_clu);
                                        if( tmpHitCluDis < MinHitDisToCluster_Neutral )
                                        {
                                                MinHitDisToCluster_Neutral = tmpHitCluDis;
                                                CluIndex = p;
                                        }
                                }
                        }

                        for(int s = 0; s < mergedclu_ch->getNumberOfElements(); s ++ )
                        {
                                Cluster* a_chcore = dynamic_cast<Cluster*>(mergedclu_ch->getElementAt(s));
                                a_cluPos = a_chcore->getPosition();
                                if( (a_cluPos - a_hitPos).Mag() < 600 )
                                {
                                        tmpHitCluDis = DisPointToBush( a_hitPos, a_chcore);
                                        if( tmpHitCluDis < MinHitDisToCluster_Charge )
                                                MinHitDisToCluster_Charge = tmpHitCluDis;
                                }
                        }

                        cout<<"MinDis "<<MinHitDisToCluster_Charge<<" : "<<MinHitDisToCluster_Neutral<<endl;

                        if(MinHitDisToCluster_Charge > MinHitDisToCluster_Neutral && CluIndex > 0 && MinHitDisToCluster_Neutral < 250)
                        {
                                FNCluEn[ finalneutralbush[CluIndex] ] += 0.11;
                        }
                }
        }

	float NAMom[3] = {0, 0, 0};
	float ScaleMMB = 0;

	for(int s = 0; s < finalneutralbush.size(); s++)
	{
		Cluster * a_clu = finalneutralbush[s];
		TVector3 BushSeedPos = a_clu->getPosition();
		ReconstructedParticleImpl * neutralparticle = new ReconstructedParticleImpl();
		float CluIEnEcal = CluEnReCal(a_clu).at(0);
		float CluIEnHcal = CluEnReCal(a_clu).at(1);
		float CluITheta = BushSeedPos.Theta()*TMath::RadToDeg();
		float CluIPhi = BushSeedPos.Phi()*TMath::RadToDeg(); 
		float CluIZ = fabs(BushSeedPos.Z());
		float ScaleETheta = 1.0;
		float ScaleEPhi = 1.0;
		float difPhi = fabs(CluIPhi/45.-int(CluIPhi/45.))*45.;
		if(CluIZ < 2350. && difPhi < 27 && difPhi > 7) ScaleEPhi = 1.0+0.03/20.*(difPhi-7.);
		
		if (fabs(fabs(CluITheta-90.)-14.) < 1) ScaleETheta = 1.0 + 0.02;
		else if (fabs(CluITheta-90.) < 48.) ScaleETheta = 1.0 - 0.02/48.*(48.-fabs(CluITheta-90.));
		else if (fabs(CluITheta-90.) < 52.5)  ScaleETheta = 1.0 + 0.06/4.5*(fabs(CluITheta-90.)-48.);
		else if (fabs(CluITheta-90.) < 77.) ScaleETheta = 1.0 + 0.0125;
		else ScaleETheta = 1.0 + 0.0125 + 0.08/13.*(fabs(CluITheta-90.)-77.);
		
		float CluFinalE = CluIEnEcal*ScaleETheta*ScaleEPhi+CluIEnHcal;


		ScaleMMB = 1.0;
		//ScaleMMB = 1.01;
		//ScaleMMB = SCale1;
		if( PhotonTag(a_clu) == 1 )
		{
			neutralparticle->setType(22);
			if(a_clu->getEnergy() < 3.0) ScaleMMB = ScaleMMB*(1-2/40.);
			else ScaleMMB = ScaleMMB*(1-1/30.);
		}
		else
		{
			neutralparticle->setType(2112);
			if( a_clu->getEnergy() > 5.0 )
			{
				//ScaleMMB = 1.2;
				ScaleMMB = 0.97; 
			}
			if(a_clu->getEnergy() >= 3.0) ScaleMMB = ScaleMMB*(1+1/40.);
		}

		if(a_clu->getEnergy() < 1.0)
		{
			//ScaleMMB = 1.3;
			ScaleMMB = 1.24*(1+0.1*(1-a_clu->getEnergy()));
		}
		else if( CluDepthMap[a_clu] > 200 )
		{
			ScaleMMB = 1.2;
			//ScaleMMB = SCale1;
		}
		float CurrCoreEn = ScaleMMB * CluFinalE;
		
		
		
		neutralparticle->setEnergy( CurrCoreEn );
		neutralparticle->setMass( 0.0 );
		neutralparticle->setCharge( 0.0 );
		NAMom[0] = CurrCoreEn*1.0/BushSeedPos.Mag()*BushSeedPos.X();
		NAMom[1] = CurrCoreEn*1.0/BushSeedPos.Mag()*BushSeedPos.Y();
		NAMom[2] = CurrCoreEn*1.0/BushSeedPos.Mag()*BushSeedPos.Z();
		neutralparticle->setMomentum( NAMom );
		ClusterImpl * a_neclu = NaiveCluImpl(a_clu);
		a_neclu->setEnergy( CurrCoreEn );	//Reset...
		neutralparticle->addCluster(a_neclu);
		mergedclu_ne->addElement(a_neclu);
		arborrecoparticle->addElement(neutralparticle);

		// cout<<"Energy Difference... "<<CurrCoreEn<<" >? "<<a_clu->getEnergy()<<endl; 

		TotalNeEn += CurrCoreEn;
	}

	evtPP->addCollection( arborrecoparticle, "ArborPFOs");
	evtPP->addCollection( mergedclu_ne, "ArborNeutral" );

	std::vector<std::string> TotalSubClass; 
	TotalSubClass.push_back("EcalBushes");
	TotalSubClass.push_back("HcalBushes");
	TotalSubClass.push_back("ArborCharged");
	TotalSubClass.push_back("ArborNeutral");
	CheckConsistence(evtPP, TotalSubClass);

	cout<<"Energy Ch/En/Total: "<<TotalChEn<<" : "<<TotalNeEn<<" : "<<TotalChEn+TotalNeEn<<endl; 

}

/*
   void BushConnect::ParticleFinalPIDReco(LCEvent * evtPP)	//ChargedPFOCreating 
   {
   }
 */

void BushConnect::processEvent( LCEvent * evtP )
{
	if (evtP)
	{

		int _eventNr = evtP->getEventNumber();

		if(_eventNr%1 == 0)
			cout<<"Nevts Processed: "<<_eventNr<<endl;

		BushConnect::Clean();	
		BushConnect::TrackSort( evtP ); 	
		BushConnect::TagCore( evtP );		//ECAL
		BushConnect::ParticleReco( evtP );

		/* create a cluster vector. 
		 *			cluster classified as charged Core. 
		 *			and neutral ones
		 *			neutral ones will be classified according to their type 
		 *			type: EM Core, Ne Core, ...,
		 *			and other properites, 
		 *			as 
		 *			std::map<Cluster, depth> ClusterDepth...
		 * 		*/

		// BushConnect::NonChargedCoreClusterMeasure( evtP );

	}
}


void BushConnect::end()
{
	std::cout<<"Bush Connection Finished, ArborObject Formed"<<std::endl;
}


